import copy
from typing import List

from calendar import monthrange

import scipy.optimize.optimize

import ghedt.peak_load_analysis_tool as plat
from scipy import interpolate
import math
import pandas as pd
import numpy as np
import pygfunction as gt


def synthetic_load_function(
        hourly_loads, avg_hourly_load, peak_load, peak_load_duration, peak_day,
        hours_in_previous_months, hours_in_month) -> None:
    # This function can be used to help generate synthetic loads. Note that the
    # lists in python (as well as dictionaries and user defined objects) are
    # passed by reference, and do not need to be returned. Therefore, the only
    # variable here that needs returned is the integer we need to update,
    # because integers, floats and strings standing alone do not fall into the
    # same category for passing by reference. They can be updated locally, and
    # the change will have no effect upon return

    hours_in_day: int = 24

    # The hourly loads for the current month are set to avg_hourly_load
    hourly_loads[hours_in_previous_months:
                 hours_in_previous_months + hours_in_month] = \
        [avg_hourly_load] * hours_in_month

    # The peak load will split the noon hour
    peak_day_start_hour = peak_day * hours_in_day + hours_in_previous_months
    peak_day_noon = peak_day_start_hour + 12
    # If the peak load is odd, then the peak load is shifted one to the left
    # set the peak_load for a duration of peak_load_duration
    peak_load_start_hour = \
        math.floor(peak_day_noon - (float(peak_load_duration) / 2.))
    hourly_loads[peak_load_start_hour:
                 peak_load_start_hour + peak_load_duration] = \
        [peak_load] * peak_load_duration

    return


def create_synthetic_doubling_load_profile(units='W', year=2019) -> tuple:
    # Create a synthetic load doubling profile: The first month has an average
    # load of 1 (Watt or kiloWatt based on units), with a peak of 2 units. Month
    # 2 has 2 units with a 2 hour peak of 4 units. These are applied to the
    # extraction loads. The rejection loads get the opposite treatment (nearly).
    # Simply reversing the order of te list of extraction loads gives different
    # peak load days, so the following is done with rejection: the first month
    # has an average load of 12 units and a peak duration of 24 units.

    # Check the units
    if units == 'W':
        scale = 1000.
    elif units == 'kW':
        scale = 1.
    else:
        raise ValueError('Units provided are not an option.')

    # Get the number of days in each month for a given year (make 0 NULL)
    days_in_month = \
        [0] + [monthrange(year, i)[1] for i in range(1, 13)]
    days_in_year = sum(days_in_month)
    hours_in_day: int = 24

    # Heat rejection in the ground occurs when buildings are in cooling
    # mode, these loads appear negative on Ground extraction loads plots
    # hourly_rejection_loads: list = [0.] * days_in_year * hours_in_day
    # Heat extraction in the ground occurs when buildings are in heating
    # mode, these loads appear positive on Ground extraction load plots
    hourly_extraction_loads: list = [0.] * days_in_year * hours_in_day
    hourly_rejection_loads: list = [0.] * days_in_year * hours_in_day

    # the peak extraction will occur on day 14
    peak_day_ext: int = 7
    # the peak rejection will occur on day 17
    peak_day_rej: int = 19

    # Store the index of the last months hours
    hours_in_previous_months: int = 0
    for i in range(1, len(days_in_month)):
        # The total number of hours in this month
        hours_in_month: int = hours_in_day * days_in_month[i]
        # The extraction monthly average load
        avg_hourly_load_ext: float = float(i) * scale
        # The rejection monthly average load
        avg_hourly_load_rej: float = float(len(days_in_month) - i)

        # The peak load occurs in day peak_day and for time peak_load_duration
        peak_load_ext: float = 2. * float(i) * scale  # The monthly peak load
        peak_load_duration_ext: int = i  # The peak load duration

        # The peak load and peak load duration for rejection
        peak_load_rej: float = 2. * float(len(days_in_month) - i) * scale
        peak_load_duration_rej: int = len(days_in_month) - i

        synthetic_load_function(
            hourly_extraction_loads, avg_hourly_load_ext, peak_load_ext,
            peak_load_duration_ext, peak_day_ext, hours_in_previous_months,
            hours_in_month)
        synthetic_load_function(
            hourly_rejection_loads, avg_hourly_load_rej, peak_load_rej,
            peak_load_duration_rej, peak_day_rej, hours_in_previous_months,
            hours_in_month)

        # Count up on extraction days from 7 to 18
        peak_day_ext += 1
        # Count down on rejection days from 18 to 7
        peak_day_rej -= 1

        # track the hours that have passed
        hours_in_previous_months += hours_in_month

    return hourly_rejection_loads, hourly_extraction_loads


class Loads:
    def __init__(self, hourly_loads: List[float], g_sts, ts, two_pi_k, Rb, year=2019):
        self.hourly = hourly_loads

        self.g_sts = g_sts
        self.ts = ts
        self.two_pi_k = two_pi_k
        self.Rb = Rb

        # Get the number of days in each month for a given year (make 0 NULL)
        self.days_in_month = \
            [0] + [monthrange(year, i)[1] for i in range(1, 13)]
        assert len(hourly_loads) == sum(self.days_in_month) * 24., \
            "The total number of hours in the year are not equal. Is this a" \
            " leap year?"

        # This block of data holds the compact monthly representation of the
        # loads. The intention is that these loads will usually repeat. It's
        # possible that for validation or design purposes, users may wish to
        # specify loads that differ from year to year. For these arrays,
        # January is the second item (1) and December the last (12)
        # We'll reserve the first item (0) for an annual total or peak
        self.monthly_total = [0.] * 13
        self.monthly_peak = [0.] * 13
        self.monthly_average = [0.] * 13
        self.monthly_peak_day = [0.] * 13
        self.split_load_by_month()

        # 48 hour loads are going to be necessary for the hourly simulation for
        # finding the peak load duration
        # These will be a 2D list, a list of 48 hour loads in each index
        # Make 0 position NULL
        self.monthly_two_day_hourly_peak = [[0]]
        self.process_two_day_loads()

        # Now we need to perform 48-hour simulations to determine the
        # monthly peak load hours
        self.monthly_two_day_fluid_temps_nm = [[0]]
        # Stores two day (48 hour) fluid temperatures with peak load
        self.monthly_two_day_fluid_temps_pk = [[0]]

        # duration of monthly peak clg load in hours
        self.monthly_peak_duration = [0] * 13
        self.find_peak_durations()

    def split_load_by_month(self):
        # Split the loads into peak, total and average loads for each month

        hours_in_day = 24
        # Store the index of the last months hours
        hours_in_previous_months = 0
        for i, days_in_month in enumerate(self.days_in_month[1:], start=1):
            hours_in_month = hours_in_day * days_in_month
            # Slice the hours in this current month
            month_slice = slice(hours_in_previous_months,
                                hours_in_previous_months + hours_in_month)
            month_loads = self.hourly[month_slice]
            assert len(month_loads) == hours_in_month

            # Total loads for the month in kW
            self.monthly_total[i] = sum(month_loads)

            # Peak load in kW
            # monthly peak cooling
            self.monthly_peak[i] = max(month_loads)

            # Average monthly load in kW
            self.monthly_average[i] = self.monthly_total[i] / len(month_loads)

            # Day of month the peak load occurs (e.g. 1-31)
            self.monthly_peak_day[i] = math.ceil(month_loads.index(
                self.monthly_peak[i]) / hours_in_day)

            hours_in_previous_months += hours_in_month

        return

    def process_two_day_loads(self) -> None:
        # The two day (48 hour) two day loads are selected by locating the day
        # the peak load of the month occurs on, and pulling a 48-hour load
        # profile -- the day before and the day of

        hours_in_day = 24
        hours_in_year = len(self.hourly)

        # Add the last day of the year to the beginning of the loads to account
        # for the possibility that a peak load occurs on the first day of the
        # year

        hourly_loads = \
            self.hourly[hours_in_year-hours_in_day:hours_in_year] + \
            self.hourly + self.hourly[0:hours_in_day]

        # Keep track of how many hours are in
        # start at 24 since we added the last day of the year to the beginning
        hours_in_previous_months = hours_in_day
        # loop over all 12 months
        for i, days_in_month in enumerate(self.days_in_month[1:], start=1):
            hours_in_month = hours_in_day * days_in_month

            monthly_peak_day = self.monthly_peak_day[i]

            # Get the starting hour of the day before the peak cooling load day
            monthly_peak_hour_start = hours_in_previous_months + (monthly_peak_day-1) * hours_in_day

            # monthly cooling loads (or heat rejection) in kWh
            two_day_hourly_peak_load = \
                hourly_loads[monthly_peak_hour_start:monthly_peak_hour_start+2*hours_in_day]

            assert len(two_day_hourly_peak_load) == 2*hours_in_day

            # Double check ourselves
            monthly_peak_day_start = \
                int((monthly_peak_hour_start-hours_in_day) / hours_in_day)
            monthly_peak_cl_hour_month = \
                int(monthly_peak_day_start - sum(self.days_in_month[0:i]))
            assert monthly_peak_cl_hour_month == monthly_peak_day - 1

            # monthly loads in kWh
            self.monthly_two_day_hourly_peak.append(two_day_hourly_peak_load)

            hours_in_previous_months += hours_in_month

        return

    def find_peak_durations(self) -> None:
        # Find the peak durations using hourly simulations for 2 days

        for i in range(1, len(self.days_in_month)):
            # Scale all the loads by the peak load
            # Perform an hourly simulation with the scaled loads
            # Perform an hourly simulation with a load of 1, or the peak loads
            # divided by the peak

            # two day cooling loads (or heat rejection) in kWh
            current_two_day_load = [0.] + self.monthly_two_day_hourly_peak[i]

            # This tolerance applies to the difference between the current
            # months peak load and the maximum of the two-day load. If the
            # absolute value of the difference between the current months
            # peak load and the current two-day peak load is within this
            # tolerance, then the maximum of the two-day load is equal to the
            # maximum of the current month. If the absolute difference is
            # greater than the tolerance, then the two-day peak load contains
            # a load greater than the current months peak load. The tolerance
            # could ONLY be exceeded when the first 24 hours is located in the
            # previous month.
            tol = 0.1

            # Ensure the peak load for the two-day load profile is the same or
            # greater than the monthly peak load. This check is done in case
            # the previous month contains a higher load than the current month.
            load_diff = self.monthly_peak[i] - max(current_two_day_load)
            # monthly peak cooling load (or heat rejection) in kW
            if abs(load_diff) < tol:
                current_month_peak_cl = self.monthly_peak[i]
            else:
                current_month_peak_cl = max(current_two_day_load)

            # monthly average cooling load (or heat rejection) in kW
            current_month_avg_cl = self.monthly_average[i]

            if current_month_peak_cl != 0.0:
                peak_duration, q_peak, q_nominal = \
                    self.perform_current_month_simulation(
                        current_two_day_load, current_month_peak_cl,
                        current_month_avg_cl, self.monthly_two_day_fluid_temps_pk,
                        self.monthly_two_day_fluid_temps_nm)
            else:
                peak_duration = 1.0e-6

            # Set the monthly cooling load duration
            self.monthly_peak_duration[i] = peak_duration

        return

    @staticmethod
    def simulate_hourly(hour_time, q, g_sts, Rb, two_pi_k, ts):
        # An hourly simulation for the fluid temperature
        # Chapter 2 of Advances in Ground Source Heat Pumps

        q_dt = np.hstack((q[1:] - q[:-1]))

        dT_fluid = [0]
        for n in range(1, len(hour_time)):
            # Take the last i elements of the reversed time array
            _time = hour_time[n] - hour_time[0:n]
            # _time = time_values_reversed[n - i:n]
            g_values = g_sts(np.log((_time * 3600.) / ts))
            # Tb = Tg + (q_dt * g)  (Equation 2.12)
            delta_Tb_i = (q_dt[0:n] / two_pi_k).dot(g_values)
            # Delta mean HPEFT fluid temperature
            Tf_mean = delta_Tb_i + q[n] * Rb
            dT_fluid.append(Tf_mean)

        return dT_fluid

    def perform_current_month_simulation(
            self, two_day_hourly_peak_load, peak_load, avg_load,
            two_day_fluid_temps_pk, two_day_fluid_temps_nm):
        g_sts = self.g_sts
        ts = self.ts
        two_pi_k = self.two_pi_k
        Rb = self.Rb

        hours_in_day = 24
        hour_time = np.array(list(range(0, 2 * hours_in_day + 1)))
        # Two day peak cooling load scaled down by average (q_max - q_avg)
        q_peak = np.array([0.] + [peak_load - avg_load] * (2 * hours_in_day))
        # Two day nominal cooling load (q_i - q_avg) / q_max * q_i
        q_nominal = np.array(
            [0.] + [(two_day_hourly_peak_load[i] - avg_load) / peak_load *
                    two_day_hourly_peak_load[i] for i in range(1, len(q_peak))])
        # Get peak fluid temperatures using peak load
        dT_fluid_pk = self.simulate_hourly(
            hour_time, q_peak, g_sts, Rb, two_pi_k, ts)
        two_day_fluid_temps_pk.append(dT_fluid_pk)
        # Get nominal fluid temperatures using nominal load
        dT_fluid_nm = self.simulate_hourly(
            hour_time, q_nominal, g_sts, Rb, two_pi_k, ts)
        two_day_fluid_temps_nm.append(dT_fluid_nm)

        dT_fluid_nm_max = max(dT_fluid_nm)

        if dT_fluid_nm_max > 0.0:
            f = scipy.interpolate.interp1d(dT_fluid_pk, hour_time)
            peak_duration = f(dT_fluid_nm_max).tolist()
        else:
            peak_duration = 1.0e-6

        return peak_duration, q_peak, q_nominal

    def peak_energy(self, m_idx):
        # gives htg load pk energy in kWh
        return self.monthly_peak[m_idx] * self.monthly_peak_duration[m_idx]

    def peak_hour_range(self, m_idx, total_months):
        # Place the peaks roughly midway through the day they occur on.
        # (In JDS's opinion, this should be amply accurate for the
        # hybrid time step.)
        # Catch the first and last peak hours to make sure they aren't 0
        # Could only be 0 when the first month has no load.
        _hours_to_mid_day = (self.monthly_peak_day[m_idx] - 1) * 24 + 12
        _half_peak_duration = (self.monthly_peak_duration[m_idx] / 2)
        first_hour_peak = \
            firstmonthhour(total_months) + _hours_to_mid_day - _half_peak_duration
        if first_hour_peak < 0.:
            first_hour_peak = 1.0e-6
        last_hour_peak = first_hour_peak + self.monthly_peak_duration[m_idx]
        return first_hour_peak, last_hour_peak


class HybridLoad:
    def __init__(self,
                 hourly_rejection_loads: list, hourly_extraction_loads: list,
                 bhe: plat.borehole_heat_exchangers.SingleUTube,
                 radial_numerical: plat.radial_numerical_borehole.RadialNumericalBH,
                 sim_params: plat.media.SimulationParameters,
                 COP_rejection=None, COP_extraction=None, year=2019):
        # Store the borehole heat exchanger
        self.bhe = bhe
        # Store the radial numerical g-function value
        # Note: this is intended to be a scipy.interp1d object
        self.radial_numerical = radial_numerical

        _g_sts = self.radial_numerical.g_sts
        _ts = self.radial_numerical.t_s
        _two_pi_k = 2. * np.pi * self.bhe.soil.k
        _Rb = self.bhe.compute_effective_borehole_resistance()

        # monthly cooling loads (or heat rejection)
        self.rejection = Loads(
            hourly_rejection_loads, _g_sts, _ts, _two_pi_k, _Rb, year=year)
        # monthly heating loads (or heat extraction)
        self.extraction = Loads(
            hourly_extraction_loads, _g_sts, _ts, _two_pi_k, _Rb, year=year)

        if COP_extraction is None:
            self.COP_extraction = 2.5  # When the building is heating mode
        else:
            self.COP_extraction = COP_extraction
        if COP_rejection is None:
            self.COP_rejection = 4.    # When the building is in cooling mode
        else:
            self.COP_rejection = COP_rejection

        # Simulation start and end month
        self.startmonth = sim_params.start_month
        self.endmonth = sim_params.end_month

        self.peakretainstart = 12  # use peak laods for first 12 months
        self.peakretainend = 12  # use peak loads for last 12 months
        # This block of data holds the sequence of loads. This is an
        # intermediate form, where the load values hold the actual loads,
        # not the the devoluted loads
        self.load = np.array(0)  # holds the load during the period
        self.hour = np.array(0)  # holds the last hour of a period
        self.sfload = np.array(0)  # holds the load in terms of step functions

        self.process_monthly_loads()

    def __repr__(self):
        output = str(self.__class__) + '\n'

        output += self.create_dataframe_of_peak_analysis().to_string()

        return output

    @staticmethod
    def split_heat_and_cool(hourly_heat_extraction, units='W'):
        """
         JCC 02.16.2020
         Split the provided loads into heating and cooling. Heating is positive,
         cooling is negative.
         :return: Loads split into heating and cooling
         """
        # Expects hourly_heat_extraction to be in Watts

        # Heat rejection in the ground occurs when buildings are in cooling
        # mode, these loads appear negative on Ground extraction loads plots
        hourly_rejection_loads: list = [0.] * len(hourly_heat_extraction)
        # Heat extraction in the ground occurs when buildings are in heating
        # mode, these loads appear positive on Ground extraction load plots
        hourly_extraction_loads: list = [0.] * len(hourly_heat_extraction)

        if units == 'W':
            scale = 1000.
        elif units == 'kW':
            scale = 1.
        else:
            raise ValueError('Units provided are not an option.')

        for i, l_hour in enumerate(hourly_heat_extraction):
            if l_hour >= 0.0:
                # Heat is extracted from ground when > 0
                hourly_extraction_loads[i] = l_hour / scale
            else:
                # Heat is rejected to ground when < 0
                hourly_rejection_loads[i] = l_hour / -scale

        return hourly_rejection_loads, hourly_extraction_loads

    def create_dataframe_of_peak_analysis(self) -> pd.DataFrame:
        # cooling loads (or heat rejection)
        cl = self.rejection
        # heating loads (or heat extraction)
        hl = self.extraction

        # The fields are: sum, peak, avg, peak day, peak duration
        hybrid_time_step_fields = {'Total': {}, 'Peak': {},
                  'Average': {}, 'Peak Day': {},
                  'Peak Duration': {}}

        d: dict = {}
        # For all the months, create dictionary of fields
        for i in range(1, 13):
            month_name = number_to_month(i)
            d[month_name] = copy.deepcopy(hybrid_time_step_fields)

            # set total
            d[month_name]['Total']['rejection'] = cl.monthly_total[i]
            d[month_name]['Total']['extraction'] = hl.monthly_total[i]
            # set peak
            d[month_name]['Peak']['rejection'] = cl.monthly_peak[i]
            d[month_name]['Peak']['extraction'] = hl.monthly_peak[i]
            # set average
            d[month_name]['Average']['rejection'] = cl.monthly_average[i]
            d[month_name]['Average']['extraction'] = hl.monthly_average[i]
            # set peak day
            d[month_name]['Peak Day']['rejection'] = cl.monthly_peak_day[i]
            d[month_name]['Peak Day']['extraction'] = hl.monthly_peak_day[i]
            # set peak duration
            d[month_name]['Peak Duration']['rejection'] = \
                cl.monthly_peak_duration[i]
            d[month_name]['Peak Duration']['extraction'] = \
                hl.monthly_peak_duration[i]

        # Convert the dictionary into a multi-indexed pandas dataframe
        arrays = [[], []]
        for field in hybrid_time_step_fields:
            arrays[0].append(field)
            arrays[0].append(field)
            arrays[1].append('rejection')
            arrays[1].append('extraction')
        tuples = list(zip(*arrays))
        index = pd.MultiIndex.from_tuples(tuples, names=['Fields', 'Load Type'])

        res = []
        for month in d:
            tmp = []
            for field in d[month]:
                for load_type in d[month][field]:
                    tmp.append(d[month][field][load_type])
            res.append(tmp)
        res = np.array(res)

        df = pd.DataFrame(res, index=list(d.keys()), columns=index)

        return df

    def process_monthly_loads(self):
        # Converts monthly load format to sequence of loads needed for
        # simulation
        # This routine is taking loads applied to the ground NOT to a heat pump.

        # cooling loads (or heat rejection)
        cl = self.rejection
        # heating loads (or heat extraction)
        hl = self.extraction

        # First, begin array with zero load before simulation starts.
        self.load = np.append(self.load, 0)
        #        self.sfload = np.append(self.sfload,0)
        lastzerohour = firstmonthhour(self.startmonth) - 1
        self.hour = np.append(self.hour, lastzerohour)

        for i in range(self.startmonth, self.endmonth + 1):
            mi = i % 12  # month index
            if mi == 0:
                mi = 12

            # set flag that determines if peak load will be included
            if i < self.startmonth + self.peakretainstart:
                include_peak = True
            elif i > self.endmonth - self.peakretainend:
                include_peak = True
            else:
                include_peak = False

            def _place_loads(order: List[Loads], sign: List[float], rate):
                for k, _ in enumerate(order):
                    if (order[k].monthly_peak[mi] > 0 and include_peak):
                        _first_hour_peak, _last_hour_peak = \
                            order[k].peak_hour_range(mi, i)
                        self.load = np.append(self.load, rate)
                        self.hour = np.append(self.hour, _first_hour_peak)
                        # peak
                        self.load = np.append(
                            self.load, sign[k] * order[k].monthly_peak[mi])
                        self.hour = np.append(self.hour, _last_hour_peak)
                # rest of month
                lastavghour = lastmonthhour(i)
                self.load = np.append(self.load, rate)
                self.hour = np.append(self.hour, lastavghour)

            # There may be a more sophisticated way to do this, but I will loop
            # through the lists mduration is the number of hours over which to
            # calculate the average value for the month (edited)
            if include_peak:
                mduration = \
                    monthdays(mi) * 24 - cl.monthly_peak_duration[mi] - \
                    hl.monthly_peak_duration[mi]
                mpeak_cl = cl.peak_energy(mi)  # peak clg energy
                mpeak_hl = hl.peak_energy(mi)  # peak htg energy
                mload = cl.monthly_total[mi] - hl.monthly_total[mi] - mpeak_cl + mpeak_hl
                mrate = mload / mduration
                peak_day_diff = cl.monthly_peak_day[mi] - hl.monthly_peak_day[mi]
            else:
                # peak load not used this month
                mduration = monthdays(mi) * 24
                mload = cl.monthly_total[mi] - hl.monthly_total[mi]
                mrate = mload / mduration
                peak_day_diff = 0

            if peak_day_diff < 0:
                order = [cl, hl]
                sign = [1.0, -1.0]
                _place_loads(order, sign, mrate)
            elif peak_day_diff > 0:
                order = [hl, cl]
                sign = [-1.0, 1.0]
                _place_loads(order, sign, mrate)
            else:
                # monthly peak heating day and cooling day are the same
                # in this case, we are ignoring the peaks
                # This is also used for the case where ipf[i] is False
                # A more sophisticated default could be use, like placing the
                # peaks on the 10th and 20th
                lastavghour = lastmonthhour(i)
                self.load = np.append(self.load, mrate)
                self.hour = np.append(self.hour, lastavghour)

        # Now fill array containing step function loads
        # Note they are paired with the ending hour, so the ith load will start
        # with the (i-1)th time
        n = self.hour.size
        # Note at this point the load and hour np arrays contain zeroes in
        # indices zero and one, then continue from there.
        for i in range(1, n):
            step_load = self.load[i] - self.load[i - 1]
            self.sfload = np.append(self.sfload, step_load)

    def hourly_load_representation(self, year=2019):

        # cooling loads (or heat rejection)
        cl = self.rejection
        # heating loads (or heat extraction)
        hl = self.extraction

        # Get the number of days in each month for a given year (make 0 NULL)
        days_in_month = \
            [0] + [monthrange(year, i)[1] for i in range(1, 13)]
        days_in_year = sum(days_in_month)
        hours_in_day: int = 24

        # Heat rejection in the ground occurs when buildings are in cooling
        # mode, these loads appear negative on Ground extraction loads plots
        # hourly_rejection_loads: list = [0.] * days_in_year * hours_in_day
        # Heat extraction in the ground occurs when buildings are in heating
        # mode, these loads appear positive on Ground extraction load plots
        hourly_extraction_loads: list = [0.] * days_in_year * hours_in_day
        hourly_rejection_loads: list = [0.] * days_in_year * hours_in_day

        # Store the index of the last months hours
        hours_in_previous_months: int = 0
        for i in range(1, len(days_in_month)):
            # The total number of hours in this month
            hours_in_month: int = hours_in_day * days_in_month[i]
            # monthly average cooling load (or heat rejection) in kW
            avg_hourly_load_rej: float = cl.monthly_average[i]
            # monthly average heating load (or heat extraction) in kW
            avg_hourly_load_ext: float = hl.monthly_average[i]

            # monthly peak heating load (or heat extraction) in kW
            peak_load_ext: float = hl.monthly_peak[i]
            # duration of monthly peak htg load in hours
            peak_load_duration_ext: int \
                = int(round(hl.monthly_peak_duration[i]))
            # day of the month on which peak htg load occurs (e.g. 1-31)
            peak_day_ext: int = hl.monthly_peak_day[i]

            peak_load_rej: float = cl.monthly_peak[i]
            peak_load_duration_rej: int \
                = int(round(cl.monthly_peak_duration[i]))
            peak_day_rej: int = cl.monthly_peak_day[i]

            synthetic_load_function(
                hourly_extraction_loads, avg_hourly_load_ext, peak_load_ext,
                peak_load_duration_ext, peak_day_ext, hours_in_previous_months,
                hours_in_month)
            synthetic_load_function(
                hourly_rejection_loads, avg_hourly_load_rej, peak_load_rej,
                peak_load_duration_rej, peak_day_rej, hours_in_previous_months,
                hours_in_month)

            # track the hours that have passed
            hours_in_previous_months += hours_in_month

        return hourly_rejection_loads, hourly_extraction_loads

    def visualize_hourly_heat_extraction(self):

        # The rejection and extraction lists will only contain loads in kW
        # that are not 0.0
        rejection: list = []
        extraction: list = []

        # cooling loads (or heat rejection)
        cl = self.rejection
        # heating loads (or heat extraction)
        hl = self.extraction

        for i in range(len(hl.hourly)):
            if self.hourly_rejection_loads[i] != 0.0:
                rejection.append((i, -hl.hourly[i]))
            if cl.hourly[i] != 0.0:
                # Flip the sign back negative because split_heat_and_cool made
                # all loads positive
                extraction.append((i, cl.hourly[i]))

        fig = gt.gfunction._initialize_figure()
        ax = fig.add_subplot(111)

        # Catch the error in cases where there are no loads
        if len(extraction) > 0:
            h_hour, h_load = zip(*extraction)
            ax.scatter(
                h_hour, h_load, c='red', s=7,
                label='Building in heating mode (ground heat extraction)')
        if len(rejection) > 0:
            c_hour, c_load = zip(*rejection)
            ax.scatter(c_hour, c_load, c='blue', s=7,
                       label='Building in cooling mode (ground heat rejection)')

        ax.set_ylabel('Ground extraction loads (kW)')
        ax.set_xlabel('Hours in one year')

        ax.grid()
        ax.set_axisbelow(True)

        fig.legend()

        fig.tight_layout(rect=(0, 0, 1, .9))

        return fig

    def visualize_fluid_simulation_results(self, month_num, extraction=True,
                                           rejection=True):
        # Plots the simulation result by month
        # rejection on ground is cooling load in building in kWh
        # extraction on ground is heating load in building in kWh

        fig = gt.gfunction._initialize_figure()
        ax = fig.add_subplot(111)

        hours_in_day = 24
        two_day_hour_time = list(range(0, 2 * hours_in_day+1))

        def draw_horizontal(two_day_fluid_temps_nm, peak_duration):
            # Find the place where the peak occurs in two day fluid temp nominal
            two_day_fluid_nm_max = max(two_day_fluid_temps_nm)
            two_day_fluid_nm_max_idx = two_day_fluid_temps_nm.index(
                two_day_fluid_nm_max)

            points = [(peak_duration, two_day_fluid_nm_max),
                      (two_day_fluid_nm_max_idx, two_day_fluid_nm_max)]
            x, y = list(zip(*points))
            ax.plot(x, y, color='orange', zorder=0)
            # Plot x marks the spot
            ax.scatter(peak_duration, two_day_fluid_nm_max, marker='X', s=50,
                       color='purple', zorder=2)
            return

        new_ticks = []

        # cooling loads (or heat rejection)
        cl = self.rejection
        # heating loads (or heat extraction)
        hl = self.extraction

        if extraction:
            # plot two day (48 hour) fluid temperatures for cooling with peak
            # load
            ax.plot(two_day_hour_time,
                    cl.monthly_two_day_fluid_temps_pk[month_num],
                    'k--', label='Peak Extraction', zorder=0)
            # plot two day (48 hour) fluid temperatures for cooling with nominal
            # load
            ax.plot(two_day_hour_time,
                    cl.monthly_two_day_fluid_temps_nm[month_num],
                    'b-', label='Extraction', zorder=1)
            # Draw horizontal line from max time to fluid sim
            draw_horizontal(cl.monthly_two_day_fluid_temps_nm[month_num],
                            cl.monthly_peak_duration[month_num])
            pk_cl_duration = cl.monthly_peak_duration[month_num]
            pk_cl_duration_rounded = round(pk_cl_duration, 1)
            new_ticks.append(pk_cl_duration_rounded)

        if rejection:
            # plot two day (48 hour) fluid temperatures for heating with peak
            # load
            ax.plot(two_day_hour_time,
                    hl.monthly_two_day_fluid_temps_pk[month_num],
                    'k.', label='Peak Rejection', zorder=0)
            # plot two day (48 hour) fluid temperatures for heating with nominal
            # load
            ax.plot(two_day_hour_time,
                    hl.monthly_two_day_fluid_temps_nm[month_num],
                    'r-.', label='Rejection', zorder=1)
            draw_horizontal(hl.monthly_two_day_fluid_temps_nm[month_num],
                            hl.monthly_peak_duration[month_num])
            # duration of monthly peak htg load in hours
            pk_hl_duration = hl.monthly_peak_duration[month_num]
            pk_hl_duration_rounded = round(pk_hl_duration, 1)
            new_ticks.append(pk_hl_duration_rounded)

        ax.set_xticks(list(ax.get_xticks()) + new_ticks)

        ax.set_xlabel('Time (hours)')
        ax.set_ylabel('Change in fluid temperature, $\Delta$T$_f$ ($\degree$C)')

        ax.set_xlim([-2, 50])

        ax.grid()
        ax.set_axisbelow(True)

        fig.tight_layout()

        fig.legend()

        return fig


def number_to_month(x):
    # Convert a numeric 1-12 to a month name
    if int(x) <= 12 and int(x) > 0:

        list_of_months = {'1': 'January', '2': 'February', '3': 'March',
                          '4': 'April', '5': 'May', '6': 'June', '7': 'July',
                          '8': 'August', '9': 'September', '10': 'October',
                          '11': 'November', '12': 'December'}

        return list_of_months[str(x)]

    else:
        print('num_to_month function error: "num=' + str(x) + '"')


def monthindex(mname):
    months = {'January': 1, 'February': 2, 'March': 3, 'April': 4, 'May': 5,
              'June': 6, 'July': 7, 'August': 8, 'September': 9, 'October': 10,
              'November': 11, 'December': 12}
    mi = months.get(mname)
    return mi


def monthdays(month):
    if month > 12:
        md = month % 12
    else:
        md = month
    ndays = [31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    monthdays = ndays[md]
    return monthdays


def firstmonthhour(month):
    fmh = 1
    if month > 1:
        for i in range(1,month):
            mi = i % 12
            fmh = fmh + 24 * monthdays(mi)
    return fmh


def lastmonthhour(month):
    lmh = 0
    for i in range(1, month + 1):
        lmh = lmh + monthdays(i) * 24
    if month == 1:
        lmh = 31 * 24
    return lmh



import copy

import numpy as np
import cv2


def scale_coordinates(coordinates, scale):
    new_coordinates = []
    for i in range(len(coordinates)):
        x, y = coordinates[i]
        x *= scale
        y *= scale
        new_coordinates.append((x, y))
    return new_coordinates


def remove_cutout(coordinates, boundary=None, remove_inside=True,
                  keep_contour=True):
    if boundary is None:
        boundary = []

    # cv2.pointPolygonTest only takes integers, so we scale by 10000 and then
    # scale back to keep precision
    scale = 10000.
    coordinates = scale_coordinates(coordinates, scale)
    boundary = scale_coordinates(boundary, scale)

    _boundary = np.array(boundary, dtype=np.uint64)

    # https://stackoverflow.com/a/50670359/11637415
    # Positive - point is inside the contour
    # Negative - point is outside the contour
    # Zero - point is on the contour

    inside_points_idx = []
    outside_points_idx = []
    boundary_points_idx = []
    for i in range(len(coordinates)):
        coordinate = coordinates[i]
        dist = cv2.pointPolygonTest(_boundary, coordinate, False)
        if dist > 0.0:
            inside_points_idx.append(i)
        elif dist < 0.0:
            outside_points_idx.append(i)
        elif dist == 0.0:
            boundary_points_idx.append(i)

    new_coordinates = []
    for i in range(len(coordinates)):
        # if we want to remove inside points and keep contour points
        if remove_inside and keep_contour:
            if i in inside_points_idx:
                continue
            else:
                new_coordinates.append(coordinates[i])
        # if we want to remove inside points and remove contour points
        elif remove_inside and not keep_contour:
            if i in inside_points_idx or i in boundary_points_idx:
                continue
            else:
                new_coordinates.append(coordinates[i])
        # if we want to keep outside points and remove contour points
        elif not remove_inside and not keep_contour:
            if i in outside_points_idx or i in boundary_points_idx:
                continue
            else:
                new_coordinates.append(coordinates[i])
        # if we want to keep outside points and keep contour points
        else:
            if i in outside_points_idx:
                continue
            else:
                new_coordinates.append(coordinates[i])

    new_coordinates = scale_coordinates(new_coordinates, 1/scale)

    return new_coordinates


def determine_largest_rectangle(property_boundary):
    x_max = 0
    y_max = 0
    for i in range(len(property_boundary)):
        x, y = property_boundary[i]
        if x > x_max:
            x_max = copy.deepcopy(x)
        if y > y_max:
            y_max = copy.deepcopy(y)

    rectangle = [[0, 0],
                 [x_max, 0],
                 [x_max, y_max],
                 [0, y_max],
                 [0, 0]]

    return rectangle



import copy
import numpy as np
import json
from matplotlib.ticker import Locator
import pickle
import warnings

# utilities.py - this module contains general utility functions. The functions
# are categorically grouped by their purpose. If any one of these categories
# becomes substantial, then a module can be created that is dedicated to
# functions of that category. Some of these functions may be more applicable
# to a different module. If that is the case, the function needs to contain a
# deprecation warning until the next major release.


# Time functions
# --------------
def Eskilson_log_times():
    # Return a list of Eskilson's original 27 dimensionless points in time
    log_time = [-8.5, -7.8, -7.2, -6.5, -5.9, -5.2, -4.5, -3.963, -3.27,
                -2.864, -2.577, -2.171, -1.884,
                -1.191, -0.497, -0.274, -0.051, 0.196, 0.419,
                0.642, 0.873, 1.112, 1.335, 1.679, 2.028, 2.275, 3.003]
    return log_time


# Spatial functions
# -----------------
def borehole_spacing(borehole, coordinates):
    # Use the distance between the first pair of coordinates as the B-spacing
    x_0, y_0 = coordinates[0]
    if len(coordinates) == 1:
        # Set the spacing to be the borehole radius if there's just one borehole
        B = copy.deepcopy(borehole.r_b)
    elif len(coordinates) > 1:
        x_1, y_1 = coordinates[1]
        B = max(borehole.r_b,
                np.sqrt((x_1 - x_0) ** 2 + (y_1 - y_0) ** 2))
    else:
        raise ValueError('The coordinates_domain needs to contain a positive'
                         'number of (x, y) pairs.')
    return B


def polygonal_area(corners):
    n = len(corners) # of corners
    area = 0.0
    for i in range(n):
        j = (i + 1) % n
        area += corners[i][0] * corners[j][1]
        area -= corners[j][0] * corners[i][1]
    area = abs(area) / 2.0
    return area

# TODO: Add `set_shank` functionality to utilities.py
# def set_shank(configuration: str, rb: float, r_in: float, r_out: float):
#     raise ValueError('This function is incomplete.')
#     if configuration == 'A':
#         a = 1
#     elif configuration == 'B':
#         a = 1
#     elif configuration == 'C':
#         a = 1
#     else:
#         raise ValueError('Only configurations A, B, or C are valid.')


def make_rectangle_perimeter(length_x, length_y, origin=(0, 0)):
    # Create an outer rectangular perimeter given an origin and side lengths.
    origin_x = origin[0]
    origin_y = origin[1]
    rectangle_perimeter = \
        [[origin_x, origin_y], [origin_x + length_x, origin_y],
         [origin_x + length_x, origin_y + length_y],
         [origin_x, origin_y + length_y]]
    return rectangle_perimeter


def number_of_boreholes(length, B, func=np.ceil):
    N = func(length / B) + 1
    return int(N)


def length_of_side(N, B):
    L = (N - 1) * B
    return L


def spacing_along_length(L, N):
    B = L / (N - 1)
    return B


# Design oriented functions
# -------------------------
def sign(x: float) -> int:
    """
    Determine the sign of a value, pronounced "sig-na"
    :param x: the input value
    :type x: float
    :return: a 1 or a -1
    """
    return int(abs(x) / x)


def check_bracket(sign_xL, sign_xR, disp=None) -> bool:
    if disp is not None:
        warnings.warn('The disp option in check_bracket will be removed in '
                      'the ghedt 0.2 release.')
    if sign_xL < 0 < sign_xR:
        # Bracketed the root
        return True
    elif sign_xR < 0 < sign_xL:
        # Bracketed the root
        return True
    else:
        # The root has not been bracketed, this method will return false.
        return False


# File input/output or file path handling functions.
# --------------------------------------------------
def js_dump(file_name, d, indent=4):
    with open(file_name + '.json', 'w') as fp:
        json.dump(d, fp, indent=indent)


def js_load(filename: str):
    with open(filename) as f_in:
        return json.load(f_in)


def create_if_not(directory):
    import os
    if not os.path.exists(directory):
        os.makedirs(directory)


def create_input_file(self, file_name='ghedt_input'):
    # Store an object in a file using pickle.
    file_handler = open(file_name + '.obj', 'wb')
    pickle.dump(self, file_handler)
    file_handler.close()

    return


def read_input_file(path_to_file):
    # Open a .obj file and return the ghedt object.
    file = open(path_to_file, 'rb')
    object_file = pickle.load(file)
    file.close()

    return object_file


# Functions related to computing statistics.
# ------------------------------------------
def compute_mpe(actual: list, predicted: list) -> float:
    """
    The following mean percentage error formula is used:
    .. math::
        MPE = \dfrac{100\%}{n}\sum_{i=0}^{n-1}\dfrac{a_t-p_t}{a_t}
    Parameters
    ----------
    actual: list
        The actual computed g-function values
    predicted: list
        The predicted g-function values
    Returns
    -------
    **mean_percent_error: float**
        The mean percentage error in percent
    """
    # the lengths of the two lists should be the same
    assert len(actual) == len(predicted)
    # create a summation variable
    summation: float = 0.
    for i in range(len(actual)):
        summation += (predicted[i] - actual[i]) / actual[i]
    mean_percent_error = summation * 100 / len(actual)
    return mean_percent_error


# Functions related to plotting.
# ------------------------------
class MinorSymLogLocator(Locator):
    """
    Dynamically find minor tick positions based on the positions of
    major ticks for a symlog scaling.
    """
    def __init__(self, linthresh, nints=10):
        """
        Ticks will be placed between the major ticks.
        The placement is linear for x between -linthresh and linthresh,
        otherwise its logarithmically. nints gives the number of
        intervals that will be bounded by the minor ticks.
        """
        self.linthresh = linthresh
        self.nintervals = nints

    def __call__(self):
        # Return the locations of the ticks
        majorlocs = self.axis.get_majorticklocs()

        if len(majorlocs) == 1:
            return self.raise_if_exceeds(np.array([]))

        # add temporary major tick locs at either end of the current range
        # to fill in minor tick gaps
        # major tick difference at lower end
        dmlower = majorlocs[1] - majorlocs[0]
        # major tick difference at upper end
        dmupper = majorlocs[-1] - majorlocs[-2]

        # add temporary major tick location at the lower end
        if majorlocs[0] != 0. and \
                ((majorlocs[0] != self.linthresh and
                  dmlower > self.linthresh) or
                 (dmlower == self.linthresh and majorlocs[0] < 0)):
            majorlocs = np.insert(majorlocs, 0, majorlocs[0]*10.)
        else:
            majorlocs = np.insert(majorlocs, 0, majorlocs[0]-self.linthresh)

        # add temporary major tick location at the upper end
        if majorlocs[-1] != 0. and \
                ((np.abs(majorlocs[-1]) != self.linthresh and
                  dmupper > self.linthresh) or
                 (dmupper == self.linthresh and majorlocs[-1] > 0)):
            majorlocs = np.append(majorlocs, majorlocs[-1]*10.)
        else:
            majorlocs = np.append(majorlocs, majorlocs[-1]+self.linthresh)

        # iterate through minor locs
        minorlocs = []

        # handle the lowest part
        for i in range(1, len(majorlocs)):
            majorstep = majorlocs[i] - majorlocs[i-1]
            if abs(majorlocs[i-1] + majorstep/2) < self.linthresh:
                ndivs = self.nintervals
            else:
                ndivs = self.nintervals - 1.

            minorstep = majorstep / ndivs
            locs = np.arange(majorlocs[i-1], majorlocs[i], minorstep)[1:]
            minorlocs.extend(locs)

        return self.raise_if_exceeds(np.array(minorlocs))

    def tick_values(self, vmin, vmax):
        raise NotImplementedError('Cannot get tick locations for a '
                                  '%s type.' % type(self))




import pygfunction as gt
from copy import deepcopy
from numpy import pi, log, sqrt
from scipy.optimize import brentq
from . import borehole_heat_exchangers
import ghedt.peak_load_analysis_tool as plat


def compute_equivalent(bhe):
    # Compute an equivalent borehole heat exchanger based on the type
    if type(bhe) == plat.borehole_heat_exchangers.SingleUTube:
        _bhe = bhe
    elif type(bhe) == plat.borehole_heat_exchangers.MultipleUTube:
        _bhe = multiple_to_single(bhe)
    elif type(bhe) == plat.borehole_heat_exchangers.CoaxialPipe:
        _bhe = coaxial_to_single(bhe)
    else:
        raise ValueError('Not an acceptable BHE.')

    return _bhe


def solve_root(x, objective_function, lower=None, upper=None,
               xtol=1.0e-6, rtol=1.0e-6, maxiter=50):
    # Vary flow rate to match the convective resistance

    # Use Brent Quadratic to find the root
    # Define a lower and upper for thermal conductivities
    if lower is None:
        lower = x / 100.
    else:
        lower = lower
    if upper is None:
        upper = x * 10.
    else:
        upper = upper
    # Check objective function upper and lower bounds to make sure the root is
    # bracketed
    minus = objective_function(lower)
    plus = objective_function(upper)
    # get signs of upper and lower thermal conductivity bounds
    kg_minus_sign = int(minus / abs(minus))
    kg_plus_sign = int(plus / abs(plus))

    # Solve the root if we can, if not, take the higher value
    if kg_plus_sign != kg_minus_sign:
        x = brentq(objective_function, lower, upper,
                   xtol=xtol, rtol=rtol, maxiter=maxiter)
    elif kg_plus_sign == -1 and kg_minus_sign == -1:
        x = upper
    elif kg_plus_sign == 1 and kg_minus_sign == 1:
        x = lower

    return x


def equivalent_single_u_tube(bhe, V_fluid, V_pipe, R_conv, R_pipe):
    # Note: BHE can be double U-tube or coaxial heat exchanger

    # Compute equivalent single U-tube geometry
    n = 2
    r_p_i_prime = sqrt(V_fluid / (n * pi))
    r_p_o_prime = sqrt((V_fluid + V_pipe) / (n * pi))
    A_s_prime = n * pi * ((r_p_i_prime * 2) ** 2)
    h_prime = 1 / (R_conv * A_s_prime)
    k_p_prime = log(r_p_o_prime / r_p_i_prime) / (2 * pi * n * R_pipe)

    # Place single u-tubes at a B-spacing
    # Total horizontal space (m)
    borehole = deepcopy(bhe.b)
    spacing = borehole.r_b * 2 - (n * r_p_o_prime * 2)
    # If the spacing is negative, then the borehole is not large enough,
    # therefore, the borehole will be increased if necessary
    if spacing <= 0.0:
        borehole.r_b -= spacing  # Add on the necessary spacing to fit
        spacing = (borehole.r_b * 2.) / 10.  # make spacing 1/10th of diameter
        borehole.r_b += spacing
    s = spacing / 3  # outer tube-to-tube shank spacing (m)
    pos = plat.media.Pipe.place_pipes(s, r_p_o_prime, 1)  # Place single u-tube pipe

    # New pipe geometry
    eps = deepcopy(bhe.pipe.eps)
    rhoCp = deepcopy(bhe.pipe.rhoCp)
    pipe = \
        plat.media.Pipe(pos, r_p_i_prime, r_p_o_prime, s, eps, k_p_prime, rhoCp)

    # Don't tie together the original and equivalent BHE's
    m_flow_borehole = deepcopy(bhe.m_flow_borehole)
    fluid = deepcopy(bhe.fluid)
    grout = deepcopy(bhe.grout)
    soil = deepcopy(bhe.soil)

    # Maintain the same mass flow rate so that the Rb/Rb* is not diverged from
    eq_single_u_tube = borehole_heat_exchangers.SingleUTube(
        m_flow_borehole, fluid, borehole, pipe, grout, soil)

    # The thermal conductivity of the pipe must now be varied such that R_fp is
    # equivalent to R_fp_prime
    def objective_pipe_conductivity(pipe_k):
        eq_single_u_tube.pipe.k = pipe_k
        eq_single_u_tube.update_thermal_resistance()
        return eq_single_u_tube.R_fp - (R_conv + R_pipe)

    # Use Brent Quadratic to find the root
    # Define a lower and upper for pipe thermal conductivities
    k_p_lower = eq_single_u_tube.pipe.k / 100.
    k_p_upper = eq_single_u_tube.pipe.k * 10.

    # Solve for the mass flow rate to make the convection values equal
    solve_root(eq_single_u_tube.pipe.k, objective_pipe_conductivity,
               lower=k_p_lower, upper=k_p_upper)

    eq_single_u_tube.update_thermal_resistance()

    return eq_single_u_tube


def u_tube_volumes(u_tube):
    # Compute volumes for U-tube geometry
    # Effective parameters
    n = int(u_tube.nPipes * 2)  # Total number of tubes
    # Total inside surface area (m^2)
    A_s = n * pi * (u_tube.r_in * 2.) ** 2
    R_conv = 1 / (u_tube.h_f * A_s)  # Convection resistance (m.K/W)
    # Volumes
    V_fluid = n * pi * (u_tube.r_in ** 2)
    V_pipe = n * pi * (u_tube.r_out ** 2) - V_fluid
    V_grout = pi * (u_tube.b.r_b ** 2) - V_pipe - V_fluid
    R_pipe = \
        log(u_tube.r_out / u_tube.r_in) / \
        (n * 2 * pi * u_tube.pipe.k)
    return V_fluid, V_pipe, R_conv, R_pipe


def concentric_tube_volumes(coaxial):
    # Unpack the radii to reduce confusion in the future
    r_in_in, r_in_out = coaxial.r_inner
    r_out_in, r_out_out = coaxial.r_outer
    # Compute volumes for concentric ghe geometry
    V_fluid = pi * ((r_in_in ** 2) + (r_out_in ** 2) -
                    (r_in_out ** 2))
    V_pipe = pi * ((r_in_out ** 2) - (r_in_in ** 2) +
                   (r_out_out ** 2) - (r_out_in ** 2))
    V_grout = pi * ((coaxial.b.r_b ** 2) - (r_out_out ** 2))
    A_s = pi * 2 * r_out_in
    R_conv = 1 / (coaxial.h_fluid_a_in * A_s)
    R_pipe = log(r_out_out / r_out_in) / (2 * pi * coaxial.pipe.k[1])
    return V_fluid, V_pipe, R_conv, R_pipe


def match_effective_borehole_resistance(tube_ref, new_tube):
    # Find the thermal conductivity that makes the borehole resistances equal

    # Define objective function for varying the grout thermal conductivity
    def objective_resistance(k_g):
        # update new tubes grout thermal conductivity and relevant parameters
        new_tube.k_g = k_g
        new_tube.grout.k = k_g
        # Update Delta-circuit thermal resistances
        # Initialize stored_coefficients
        Rb_prime = new_tube.update_thermal_resistance(m_flow_borehole=None,
                                                      fluid=None)
        Rb = tube_ref.compute_effective_borehole_resistance()
        return Rb - Rb_prime

    # Use Brent Quadratic to find the root
    # Define a lower and upper for thermal conductivities
    kg_lower = 1e-02
    kg_upper = 7.
    k_g = solve_root(
        new_tube.grout.k, objective_resistance, lower=kg_lower, upper=kg_upper)
    # Ensure the grout thermal conductivity is updated
    new_tube.k_g = k_g
    new_tube.grout.k = k_g
    new_tube.update_thermal_resistance()

    return


def multiple_to_single(multiple_u_tube):
    # Find an equivalent single U-tube given multiple U-tube geometry

    # Get effective parameters for the multiple u-tube
    V_fluid, V_pipe, R_conv, R_pipe = u_tube_volumes(multiple_u_tube)

    single_u_tube = equivalent_single_u_tube(
        multiple_u_tube, V_fluid, V_pipe, R_conv, R_pipe)

    # Vary grout thermal conductivity to match effective borehole thermal
    # resistance
    match_effective_borehole_resistance(multiple_u_tube, single_u_tube)

    return single_u_tube


def coaxial_to_single(coaxial_tube):
    # Find an equivalent single U-tube given a coaxial heat exchanger
    V_fluid, V_pipe, R_conv, R_pipe = concentric_tube_volumes(coaxial_tube)

    single_u_tube = equivalent_single_u_tube(
        coaxial_tube, V_fluid, V_pipe, R_conv, R_pipe)

    # Vary grout thermal conductivity to match effective borehole thermal
    # resistance
    match_effective_borehole_resistance(coaxial_tube, single_u_tube)

    return single_u_tube



import ghedt as dt
import ghedt.peak_load_analysis_tool as plat
import pygfunction as gt
import numpy as np
import textwrap


# Common design interface
class Design:
    def __init__(self, V_flow: float, borehole: gt.boreholes.Borehole,
                 bhe_object: plat.borehole_heat_exchangers,
                 fluid: gt.media.Fluid, pipe: plat.media.Pipe,
                 grout: plat.media.Grout, soil: plat.media.Soil,
                 sim_params: plat.media.SimulationParameters,
                 geometric_constraints: dt.media.GeometricConstraints,
                 hourly_extraction_ground_loads: list, method: str = 'hybrid',
                 routine: str = 'near-square', flow: str = 'borehole'):
        self.V_flow = V_flow  # volumetric flow rate, m3/s
        self.borehole = borehole
        self.bhe_object = bhe_object  # a borehole heat exchanger object
        self.fluid = fluid  # a fluid object
        self.pipe = pipe
        self.grout = grout
        self.soil = soil
        self.sim_params = sim_params
        self.geometric_constraints = geometric_constraints
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.method = method
        if self.method == 'hourly':
            msg = 'Note: It is not recommended to perform a field selection ' \
                  'with the hourly simulation due to computation time. If ' \
                  'the goal is to validate the selected field with the ' \
                  'hourly simulation, the better solution is to utilize the ' \
                  'hybrid simulation to automatically select the field. Then ' \
                  'perform a sizing routine on the selected GHE with the ' \
                  'hourly simulation.'
            # Wrap the text to a 50 char line width and print it
            wrapper = textwrap.TextWrapper(width=72)
            word_list = wrapper.wrap(text=msg)
            for element in word_list:
                print(element)
            print('\n')

        # Check the routine parameter
        self.routine = routine
        available_routines = ['near-square', 'rectangle', 'bi-rectangle',
                              'bi-zoned']
        self.geometric_constraints.check_inputs(self.routine)
        gc = self.geometric_constraints
        if routine in available_routines:
            # If a near-square design routine is requested, then we go from a
            # 1x1 to 32x32 at the B-spacing
            # The lower end of the near-square routine is always 1 borehole.
            # There would never be a time that a user would __need__ to give a
            # different lower range. The upper number of boreholes range is
            # calculated based on the spacing and length provided.
            if routine == 'near-square':
                number_of_boreholes = \
                    dt.utilities.number_of_boreholes(
                        gc.length, gc.B, func=np.floor)
                self.coordinates_domain = \
                    dt.domains.square_and_near_square(
                        1, number_of_boreholes, self.geometric_constraints.B)
            elif routine == 'rectangle':
                self.coordinates_domain = dt.domains.rectangular(
                    gc.length, gc.width, gc.B_min, gc.B_max_x, disp=False
                )
            elif routine == 'bi-rectangle':
                self.coordinates_domain_nested = dt.domains.bi_rectangle_nested(
                    gc.length, gc.width, gc.B_min, gc.B_max_x, gc.B_max_y,
                    disp=False)
            elif routine == 'bi-zoned':
                self.coordinates_domain_nested = \
                    dt.domains.bi_rectangle_zoned_nested(
                        gc.length, gc.width, gc.B_min, gc.B_max_x, gc.B_max_y)
        else:
            raise ValueError('The requested routine is not available. '
                             'The currently available routines are: '
                             '`near-square`.')
        self.flow = flow

    def find_design(self, disp=False):
        if disp:
            title = 'Find {}...'.format(self.routine)
            print(title + '\n' + len(title) * '=')
        # Find near-square
        if self.routine == 'near-square':
            bisection_search = dt.search_routines.Bisection1D(
                self.coordinates_domain, self.V_flow, self.borehole,
                self.bhe_object, self.fluid, self.pipe, self.grout,
                self.soil, self.sim_params, self.hourly_extraction_ground_loads,
                method=self.method, flow=self.flow, disp=disp)
        # Find a rectangle
        elif self.routine == 'rectangle':
            bisection_search = dt.search_routines.Bisection1D(
                self.coordinates_domain, self.V_flow, self.borehole,
                self.bhe_object, self.fluid, self.pipe, self.grout, self.soil,
                self.sim_params, self.hourly_extraction_ground_loads,
                method=self.method, flow=self.flow, disp=disp)
        # Find a bi-rectangle
        elif self.routine == 'bi-rectangle':
            bisection_search = dt.search_routines.Bisection2D(
                self.coordinates_domain_nested, self.V_flow,
                self.borehole, self.bhe_object, self.fluid, self.pipe,
                self.grout, self.soil, self.sim_params,
                self.hourly_extraction_ground_loads, method=self.method,
                flow=self.flow, disp=disp)
        # Find bi-zoned rectangle
        elif self.routine == 'bi-zoned':
            bisection_search = dt.search_routines.BisectionZD(
                self.coordinates_domain_nested, self.V_flow, self.borehole,
                self.bhe_object, self.fluid, self.pipe, self.grout, self.soil,
                self.sim_params, self.hourly_extraction_ground_loads,
                method=self.method, flow=self.flow, disp=disp)
        else:
            raise ValueError('The requested routine is not available. '
                             'The currently available routines are: '
                             '`near-square`.')

        return bisection_search



import ghedt.peak_load_analysis_tool as plat
import numpy as np
from math import log, sqrt, exp
from math import pi
from scipy.interpolate import interp1d
from scipy.linalg.lapack import dgtsv

# Note: This was originally pulled out of Matt Mitchell's GLHE repository,
# https://github.com/mitchute/GLHE. The file has been heavily edited. Namely,
# this object takes in a borehole heat exchanger object from plat. This code
# has also been vectorized, which resulted in a 20x speed increase. It is likely
# that Mitchell did not have the code vectorized using numpy because the
# implementation was going to be placed in C++. Upon vectorizing the code, the
# three tests ran all had root mean square errors less than 1.0e-10 difference.
# The MIT License for Matt Mitchell remains, out of respect for his time and
# effort placed into the the initial implementation. - Jack C. Cook 2021.09.10

"""
MIT License

Copyright (c) 2018 Matt Mitchell

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# Time constants
DAYS_IN_YEAR = 365
HOURS_IN_DAY = 24
MIN_IN_HOUR = 60
SEC_IN_MIN = 60
SEC_IN_HOUR = SEC_IN_MIN * MIN_IN_HOUR
SEC_IN_DAY = SEC_IN_HOUR * HOURS_IN_DAY
SEC_IN_YEAR = SEC_IN_DAY * DAYS_IN_YEAR

gamma_const = 1.781072  # exp(Euler's constant)


class RadialCellType(object):
    FLUID = 1
    CONVECTION = 2
    PIPE = 3
    GROUT = 4
    SOIL = 5


class RadialNumericalBH(object):
    """
     X. Xu and Jeffrey D. Spitler. 2006. 'Modeling of Vertical Ground Loop Heat
     Exchangers with Variable Convective Resistance and Thermal Mass of the
     Fluid.' in Proceedings of the 10th International Conference on Thermal
     Energy Storage-EcoStock. Pomona, NJ, May 31-June 2.
    """

    def __init__(
            self, single_u_tube: plat.borehole_heat_exchangers.SingleUTube,
            ground_init_temp: float = 20., dtype: np.dtype = np.double):
        self.single_u_tube = single_u_tube
        self.dtype = dtype

        self.cell_inputs = {'type': 0,
                            'inner-radius': 1,
                            'center-radius': 2,
                            'outer-radius': 3,
                            'thickness': 4,
                            'conductivity': 5,
                            'heat-capacity': 6,
                            'initial-temperature': 7,
                            'volume': 8}

        # "The one dimensional model has a fluid core, an equivalent convective
        # resistance layer, a tube layer, a grout layer and is surrounded by the
        # ground."

        # cell numbers
        self.num_fluid_cells = 3
        self.num_conv_cells = 1
        self.num_pipe_cells = 4
        self.num_grout_cells = 27
        self.num_soil_cells = 500

        self.num_cells = \
            self.num_fluid_cells + self.num_conv_cells + \
            self.num_fluid_cells + self.num_grout_cells + \
            self.num_soil_cells + 1

        # Geometry and gridding procedure

        # far-field radius is set to 10m; the soil region is represented by
        # 500 cells
        far_field_radius = 10  # soil radius (in meters)
        self.r_far_field = far_field_radius - single_u_tube.b.r_b

        # borehole radius is set to the actual radius of the borehole
        self.r_borehole = single_u_tube.b.r_b

        # outer tube radius is set to sqrt(2) * r_p_o, tube region has 4 cells
        self.r_out_tube = sqrt(2) * single_u_tube.pipe.r_out

        # inner tube radius is set to r_out_tube-t_p
        self.thickness_pipe = (single_u_tube.pipe.r_out - single_u_tube.pipe.r_in)
        self.r_in_tube = self.r_out_tube - self.thickness_pipe

        # r_in_convection is set to r_in_tube - 1/4 * t
        self.r_in_convection = self.r_in_tube - self.thickness_pipe / 4.

        # r_fluid is set to r_in_convection - 3/4 * t
        self.r_fluid = self.r_in_convection - (3. / 4. * self.thickness_pipe)

        # Thicknesses of the grid regions
        self.thickness_soil = (self.r_far_field - self.r_borehole) / self.num_soil_cells
        self.thickness_grout = (self.r_borehole - self.r_out_tube) / self.num_grout_cells
        # pipe thickness is equivalent to original tube thickness
        self.thickness_conv = (self.r_in_tube - self.r_in_convection) / self.num_conv_cells
        self.thickness_fluid = (self.r_in_convection - self.r_fluid) / self.num_fluid_cells

        # other
        self.init_temp = ground_init_temp

        # other
        self.g = np.array([], dtype=self.dtype)
        self.lntts = np.array([], dtype=self.dtype)
        self.c_0 = 2. * pi * single_u_tube.soil.k
        soil_diffusivity = single_u_tube.k_s / single_u_tube.soil.rhoCp
        self.t_s = single_u_tube.b.H ** 2 / (9 * soil_diffusivity)
        # default is at least 49 hours, or up to -8.6 log time
        self.calc_time_in_sec = max([self.t_s * exp(-8.6), 49. * 3600.])
        self.g_sts = None

    def fill_radial_cell(self, radial_cell, R_p_eq, R_f_eq, R_TG_eq):

        num_fluid_cells = self.num_fluid_cells
        num_conv_cells = self.num_conv_cells
        num_pipe_cells = self.num_pipe_cells
        num_grout_cells = self.num_grout_cells
        num_soil_cells = self.num_soil_cells

        cell_summation = 0

        # load fluid cells
        for idx in range(cell_summation, num_fluid_cells+cell_summation):
            cell_type = RadialCellType.FLUID
            thickness = self.thickness_fluid
            center_radius = self.r_fluid + idx * thickness

            if idx == 0:
                inner_radius = center_radius
            else:
                inner_radius = center_radius - thickness / 2.0

            outer_radius = center_radius + thickness / 2.0

            # The equivalent thermal mass of the fluid can be calculated from
            # equation (2)
            # pi (r_in_conv ** 2 - r_f **2) C_eq_f = 2pi r_p_in**2 * C_f
            rho_cp_eq = (2. * (self.single_u_tube.pipe.r_in**2) * self.single_u_tube.fluid.rho * self.single_u_tube.fluid.cp) / ((self.r_in_convection**2) - (self.r_fluid**2)  )
            k_eq = rho_cp_eq / self.single_u_tube.fluid.cp

            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [cell_type, inner_radius, center_radius, outer_radius,
                 thickness, k_eq, rho_cp_eq, self.init_temp, volume],
                dtype=self.dtype)
        cell_summation += num_fluid_cells
        assert cell_summation == num_fluid_cells

        # load convection cells
        for idx in range(cell_summation, num_conv_cells+cell_summation):
            j = idx - cell_summation
            cell_type = RadialCellType.CONVECTION
            thickness = self.thickness_conv
            inner_radius = self.r_in_convection + j * thickness
            center_radius = inner_radius + thickness / 2.0
            outer_radius = inner_radius + thickness
            k_eq = log(self.r_in_tube / self.r_in_convection) / (2. * pi * R_f_eq)
            rho_cp = 1.
            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [cell_type, inner_radius, center_radius, outer_radius,
                 thickness, k_eq, rho_cp, self.init_temp, volume],
                dtype=self.dtype)
        cell_summation += num_conv_cells
        assert cell_summation == (num_fluid_cells + num_conv_cells)

        # load pipe cells
        for idx in range(cell_summation, num_pipe_cells+cell_summation):
            j = idx - cell_summation
            cell_type = RadialCellType.PIPE
            thickness = self.thickness_pipe
            inner_radius = self.r_in_tube + j * thickness
            center_radius = inner_radius + thickness / 2.0
            outer_radius = inner_radius + thickness
            conductivity = log(self.r_borehole / self.r_in_tube) / (
                        2. * pi * R_p_eq)
            rho_cp = self.single_u_tube.pipe.rhoCp
            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [cell_type, inner_radius, center_radius, outer_radius,
                 thickness, conductivity, rho_cp, self.init_temp, volume],
                dtype=self.dtype)
        cell_summation += num_pipe_cells
        assert cell_summation == \
               (num_fluid_cells + num_conv_cells + num_pipe_cells)

        # load grout cells
        for idx in range(cell_summation, num_grout_cells+cell_summation):
            j = idx - cell_summation
            cell_type = RadialCellType.GROUT
            thickness = self.thickness_grout
            inner_radius = self.r_out_tube + j * thickness
            center_radius = inner_radius + thickness / 2.0
            outer_radius = inner_radius + thickness
            conductivity = log(self.r_borehole / self.r_in_tube) / (
                        2. * pi * R_TG_eq)
            rho_cp = self.single_u_tube.grout.rhoCp
            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [cell_type, inner_radius, center_radius, outer_radius,
                 thickness, conductivity, rho_cp, self.init_temp, volume],
                dtype=self.dtype)
        cell_summation += num_grout_cells
        assert cell_summation == \
               (num_fluid_cells + num_conv_cells + num_pipe_cells +
                num_grout_cells)

        # load soil cells
        for idx in range(cell_summation, num_soil_cells+cell_summation):
            j = idx - cell_summation
            cell_type = RadialCellType.SOIL
            thickness = self.thickness_soil
            inner_radius = self.r_borehole + j * thickness
            center_radius = inner_radius + thickness / 2.0
            outer_radius = inner_radius + thickness
            conductivity = self.single_u_tube.soil.k
            rho_cp = self.single_u_tube.soil.rhoCp
            volume = pi * (outer_radius ** 2 - inner_radius ** 2)
            radial_cell[:, idx] = np.array(
                [cell_type, inner_radius, center_radius, outer_radius,
                 thickness, conductivity, rho_cp, self.init_temp, volume],
                dtype=self.dtype)
        cell_summation += num_soil_cells

    def calc_sts_g_functions(
            self, single_u_tube, final_time=None, calculate_at_bh_wall=False) -> tuple:

        self.__init__(single_u_tube)

        Rb = self.single_u_tube.compute_effective_borehole_resistance()

        R_f_eq = self.single_u_tube.R_f / 2.
        R_p_eq = self.single_u_tube.R_p / 2.
        R_TG_eq = Rb - R_f_eq

        # Pass radial cell by reference and fill here so that it can be
        # destroyed when this method returns
        radial_cell = np.zeros(shape=(len(self.cell_inputs), self.num_cells),
                               dtype=self.dtype)
        self.fill_radial_cell(radial_cell, R_p_eq, R_f_eq, R_TG_eq)

        if final_time is None:
            final_time = self.calc_time_in_sec

        g = []
        lntts = []

        _dl = np.zeros(self.num_cells-1)
        _d = np.zeros(self.num_cells)
        _du = np.zeros(self.num_cells-1)
        _b = np.zeros(self.num_cells)

        heat_flux = 1.
        init_temp = self.init_temp

        time = (1e-12) - 120
        time_step = 120

        type_idx = self.cell_inputs['type']
        r_in_idx = self.cell_inputs['inner-radius']
        r_center_idx = self.cell_inputs['center-radius']
        r_out_idx = self.cell_inputs['outer-radius']
        thickness_idx = self.cell_inputs['thickness']
        k_idx = self.cell_inputs['conductivity']
        rhoCp_idx = self.cell_inputs['heat-capacity']
        temperature_idx = self.cell_inputs['initial-temperature']
        previous_temperature_idx = self.cell_inputs['initial-temperature']
        volume_idx = self.cell_inputs['volume']

        _fe_1 = np.zeros(shape=(self.num_cells - 2), dtype=self.dtype)
        _fe_2 = np.zeros_like(_fe_1)
        _ae = np.zeros_like(_fe_2)
        _fw_1 = np.zeros_like(_ae)
        _fw_2 = np.zeros_like(_fw_1)
        _aw = np.zeros_like(_fw_2)
        _ad = np.zeros_like(_aw)

        _west_cell = radial_cell[:, 0:self.num_cells - 2]
        _center_cell = radial_cell[:, 1:self.num_cells - 1]
        _east_cell = radial_cell[:, 2:self.num_cells - 0]

        while True:

            time += time_step

            # For the idx == 0 case:

            fe_1 = \
                log(radial_cell[r_out_idx, 0] / radial_cell[r_center_idx, 0]) \
                / (2. * pi * radial_cell[k_idx, 0])
            fe_2 = \
                log(radial_cell[r_center_idx, 1] / radial_cell[r_in_idx, 1]) \
                / (2. * pi * radial_cell[k_idx, 1])
            ae = 1 / (fe_1 + fe_2)

            ad = \
                radial_cell[rhoCp_idx, 0] * radial_cell[volume_idx, 0] / time_step

            _d[0] = -ae / ad - 1
            _du[0] = ae / ad
            _b[0] = -radial_cell[previous_temperature_idx, 0] - heat_flux / ad

            # For the idx == n-1 case

            _dl[self.num_cells-2] = 0.
            _d[self.num_cells-1] = 1.
            _b[self.num_cells-1] = \
                radial_cell[previous_temperature_idx, self.num_cells-1]

            # Now handle the 1 to n-2 cases with numpy slicing and vectorization

            def fill_f1(fx_1, cell):
                fx_1[:] = np.log(cell[r_out_idx, :] / cell[r_center_idx, :]) \
                          / (2. * pi * cell[k_idx, :])
                return

            def fill_f2(fx_2, cell):
                fx_2[:] = np.log(cell[r_center_idx, :] / cell[r_in_idx, :]) \
                          / (2. * pi * cell[k_idx, :])
                return

            fill_f1(_fe_1, _center_cell)
            fill_f2(_fe_2, _east_cell)
            _ae[:] = 1. / (_fe_1 + _fe_2)

            fill_f1(_fw_1, _west_cell)
            fill_f2(_fw_2, _center_cell)
            _aw[:] = -1. / (_fw_1 + _fw_2)

            _ad[:] = _center_cell[rhoCp_idx, :] * _center_cell[volume_idx, :] / time_step

            _dl[0:self.num_cells-2] = -_aw / _ad
            _d[1:self.num_cells-1] = _aw / _ad - _ae / _ad - 1.
            _du[1:self.num_cells-1] = _ae / _ad
            _b[1:self.num_cells-1] = -radial_cell[previous_temperature_idx, 1:self.num_cells-1]

            # Tri-diagonal matrix solver
            # High level interface to LAPACK routine
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.linalg.lapack.dgtsv.html#scipy.linalg.lapack.dgtsv
            dgtsv(_dl, _d, _du, _b, overwrite_b=1)

            radial_cell[previous_temperature_idx, :] = _b
            radial_cell[temperature_idx, :] = _b

            if calculate_at_bh_wall:
                raise ValueError(
                    'This portion of the code does not currently run with this '
                    'vectorized radial numerical implementation.')
                bh_wall_temp = 0

                # calculate bh wall temp
                for idx, this_cell in enumerate(self.cells):
                    west_cell = self.cells[idx]
                    east_cell = self.cells[idx + 1]

                    if west_cell.type == RadialCellType.GROUT and east_cell.type == RadialCellType.SOIL:
                        west_conductance_num = 2 * pi * west_cell.conductivity
                        west_conductance_den = log(west_cell.outer_radius / west_cell.inner_radius)
                        west_conductance = west_conductance_num / west_conductance_den

                        east_conductance_num = 2 * pi * east_cell.conductivity
                        east_conductance_den = log(east_cell.center_radius / west_cell.inner_radius)
                        east_conductance = east_conductance_num / east_conductance_den

                        bh_wall_temp_num_1 = west_conductance * west_cell.temperature
                        bh_wall_temp_num_2 = east_conductance * east_cell.temperature
                        bh_wall_temp_num = bh_wall_temp_num_1 + bh_wall_temp_num_2
                        bh_wall_temp_den = west_conductance + east_conductance
                        bh_wall_temp = bh_wall_temp_num / bh_wall_temp_den

                        break

                g.append(self.c_0 * ((bh_wall_temp - init_temp) / heat_flux))
            else:
                g.append(self.c_0 * ((radial_cell[temperature_idx, 0] - init_temp) / heat_flux - Rb))

            lntts.append(log(time / self.t_s))

            if time >= final_time - time_step:
                self.g = np.insert(self.g, 0, g, axis=0)
                self.lntts = np.insert(self.lntts, 0, lntts, axis=0)
                break

        self.g_sts = interp1d(lntts, g)

        return self.lntts, self.g



import copy
import unittest
import os

import ghedt as dt
import ghedt.peak_load_analysis_tool as plat
import pygfunction as gt

import pandas as pd


TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__),
                                 'Atlanta_Office_Building_Loads.csv')


class DesignBase:
    def __init__(self):
        # Borehole dimensions
        # -------------------
        H = 96.  # Borehole length (m)
        D = 2.  # Borehole buried depth (m)
        r_b = 0.075  # Borehole radius (m)

        # Pipe dimensions
        # ---------------
        # Single and Multiple U-tubes
        r_out = 26.67 / 1000. / 2.  # Pipe outer radius (m)
        r_in = 21.6 / 1000. / 2.  # Pipe inner radius (m)
        s = 32.3 / 1000.  # Inner-tube to inner-tube Shank spacing (m)
        epsilon = 1.0e-6  # Pipe roughness (m)
        # Coaxial tube
        r_in_in = 44.2 / 1000. / 2.
        r_in_out = 50. / 1000. / 2.
        # Outer pipe radii
        r_out_in = 97.4 / 1000. / 2.
        r_out_out = 110. / 1000. / 2.
        # Pipe radii
        # Note: This convention is different from pygfunction
        r_inner = [r_in_in,
                   r_in_out]  # The radii of the inner pipe from in to out
        r_outer = [r_out_in,
                   r_out_out]  # The radii of the outer pipe from in to out

        # Pipe positions
        # --------------
        # Single U-tube [(x_in, y_in), (x_out, y_out)]
        pos_single = plat.media.Pipe.place_pipes(s, r_out, 1)
        # Single U-tube BHE object
        self.single_u_tube = plat.borehole_heat_exchangers.SingleUTube
        # Double U-tube
        pos_double = plat.media.Pipe.place_pipes(s, r_out, 2)
        double_u_tube = plat.borehole_heat_exchangers.MultipleUTube
        # Coaxial tube
        pos_coaxial = (0, 0)
        coaxial_tube = plat.borehole_heat_exchangers.CoaxialPipe

        # Thermal conductivities
        # ----------------------
        k_p = 0.4  # Pipe thermal conductivity (W/m.K)
        k_p_coax = [0.4, 0.4]  # Pipes thermal conductivity (W/m.K)
        k_s = 2.0  # Ground thermal conductivity (W/m.K)
        k_g = 1.0  # Grout thermal conductivity (W/m.K)

        # Volumetric heat capacities
        # --------------------------
        rhoCp_p = 1542. * 1000.  # Pipe volumetric heat capacity (J/K.m3)
        rhoCp_s = 2343.493 * 1000.  # Soil volumetric heat capacity (J/K.m3)
        rhoCp_g = 3901. * 1000.  # Grout volumetric heat capacity (J/K.m3)

        # Thermal properties
        # ------------------
        # Pipe
        self.pipe_single = \
            plat.media.Pipe(pos_single, r_in, r_out, s, epsilon, k_p, rhoCp_p)
        pipe_double = \
            plat.media.Pipe(pos_double, r_in, r_out, s, epsilon, k_p, rhoCp_p)
        pipe_coaxial = \
            plat.media.Pipe(pos_coaxial, r_inner, r_outer, 0, epsilon, k_p_coax,
                            rhoCp_p)
        # Soil
        ugt = 18.3  # Undisturbed ground temperature (degrees Celsius)
        self.soil = plat.media.Soil(k_s, rhoCp_s, ugt)
        # Grout
        self.grout = plat.media.Grout(k_g, rhoCp_g)

        # Inputs related to fluid
        # -----------------------
        # Fluid properties
        mixer = 'MEG'  # Ethylene glycol mixed with water
        percent = 0.  # Percentage of ethylene glycol added in
        self.fluid = gt.media.Fluid(mixer, percent)

        # Fluid flow rate
        V_flow = 0.2  # Borehole volumetric flow rate (L/s)
        self.V_flow_borehole = copy.deepcopy(V_flow)

        # Define a borehole
        self.borehole = gt.boreholes.Borehole(H, D, r_b, x=0., y=0.)

        # Simulation parameters
        # ---------------------
        # Simulation start month and end month
        start_month = 1
        n_years = 20
        end_month = n_years * 12
        # Maximum and minimum allowable fluid temperatures
        max_EFT_allowable = 35  # degrees Celsius
        min_EFT_allowable = 5  # degrees Celsius
        # Maximum and minimum allowable heights
        max_Height = 135.  # in meters
        min_Height = 60  # in meters
        self.sim_params = plat.media.SimulationParameters(
            start_month, end_month, max_EFT_allowable, min_EFT_allowable,
            max_Height, min_Height)


class TestNearSquare(unittest.TestCase, DesignBase):

    def setUp(self) -> None:

        DesignBase.__init__(self)
        # Note: Based on these inputs, the resulting near-square test will
        # determine a system with 156 boreholes.
        self.V_flow_system = self.V_flow_borehole * 156.

        # Process loads from file
        # -----------------------
        # read in the csv file and convert the loads to a list of length 8760
        hourly_extraction: dict = \
            pd.read_csv(TESTDATA_FILENAME).to_dict('list')
        # Take only the first column in the dictionary
        self.hourly_extraction_ground_loads: list = \
            hourly_extraction[list(hourly_extraction.keys())[0]]

        # Geometric constraints for the `near-square` routine
        # Required geometric constraints for the uniform rectangle design: B
        B = 5.  # Borehole spacing (m)
        number_of_boreholes = 32
        length = dt.utilities.length_of_side(number_of_boreholes, B)
        self.geometric_constraints = \
            dt.media.GeometricConstraints(B=B, length=length)

    def test_design_selection(self):
        # Single U-tube
        # -------------
        # Design a single U-tube with a system volumetric flow rate
        design_single_u_tube_a = dt.design.Design(
            self.V_flow_system, self.borehole, self.single_u_tube, self.fluid,
            self.pipe_single, self.grout, self.soil, self.sim_params,
            self.geometric_constraints, self.hourly_extraction_ground_loads,
            flow='system', routine='near-square')
        # Find the near-square design for a single U-tube and size it.
        bisection_search = design_single_u_tube_a.find_design()
        bisection_search.ghe.compute_g_functions()
        bisection_search.ghe.size(method='hybrid')
        H_single_u_tube_a = bisection_search.ghe.bhe.b.H

        design_single_u_tube_b = dt.design.Design(
            self.V_flow_borehole, self.borehole, self.single_u_tube, self.fluid,
            self.pipe_single, self.grout, self.soil, self.sim_params,
            self.geometric_constraints, self.hourly_extraction_ground_loads,
            flow='borehole', routine='near-square')
        # Find the near-square design for a single U-tube and size it.
        bisection_search = design_single_u_tube_b.find_design()
        bisection_search.ghe.compute_g_functions()
        bisection_search.ghe.size(method='hybrid')
        H_single_u_tube_b = bisection_search.ghe.bhe.b.H

        # Verify that the `flow` toggle is properly working
        self.assertAlmostEqual(H_single_u_tube_a, H_single_u_tube_b, places=8)
        # Verify that the proper height as been found
        # Note: This reference was calculated on MacOS. It seems that on Linux
        # the values are not equal starting around the 9th decimal place.
        H_reference = 130.39516965123445
        self.assertAlmostEqual(H_reference, H_single_u_tube_a, places=6)



import copy
import warnings

from scipy.interpolate import interp1d, lagrange
import pygfunction as gt
import numpy as np


def calculate_g_function(
        m_flow_borehole, bhe_object, time_values, coordinates, borehole,
        fluid, pipe, grout, soil, nSegments=8, end_length_ratio=0.02,
        segments='unequal', solver='equivalent', boundary='MIFT',
        segment_ratios=None, disp=False):

    boreField = []
    BHEs = []

    H = copy.deepcopy(borehole.H)
    r_b = copy.deepcopy(borehole.r_b)
    D = copy.deepcopy(borehole.D)
    tilt = copy.deepcopy(borehole.tilt)
    orientation = copy.deepcopy(borehole.orientation)

    for i in range(len(coordinates)):
        x, y = coordinates[i]
        _borehole = gt.boreholes.Borehole(H, D, r_b, x, y, tilt, orientation)
        boreField.append(_borehole)
        # Initialize pipe model
        if boundary == 'MIFT':
            bhe = \
                bhe_object(m_flow_borehole, fluid, _borehole, pipe, grout, soil)
            BHEs.append(bhe)

    alpha = soil.k / soil.rhoCp

    # setup options
    segments = segments.lower()
    if segments == 'equal':
        options = {'nSegments': nSegments, 'disp': disp}
    elif segments == 'unequal':
        if segment_ratios is None:
            segment_ratios = gt.utilities.segment_ratios(
                nSegments, end_length_ratio=end_length_ratio)
        else:
            segment_ratios = segment_ratios
        options = {'nSegments': nSegments, 'segment_ratios': segment_ratios,
                   'disp': disp}
    else:
        raise ValueError('Equal or Unequal are acceptable options '
                         'for segments.')

    if boundary == 'UHTR' or boundary == 'UBWT':
        gfunc = gt.gfunction.gFunction(
            boreField, alpha, time=time_values, boundary_condition=boundary,
            options=options, method=solver
        )
    elif boundary == 'MIFT':
        m_flow_network = len(boreField) * m_flow_borehole
        network = gt.networks.Network(
            boreField, BHEs, m_flow_network=m_flow_network, cp_f=fluid.cp)
        gfunc = gt.gfunction.gFunction(
            network, alpha, time=time_values,
            boundary_condition=boundary, options=options, method=solver)
    else:
        raise ValueError('UHTR, UBWT or MIFT are accepted boundary conditions.')

    return gfunc


def compute_live_g_function(
        B: float, H_values: list, r_b_values: list, D_values: list,
        m_flow_borehole, bhe_object, log_time,  coordinates,
        fluid, pipe, grout, soil, nSegments=8, segments='unequal',
        solver='equivalent', boundary='MIFT', segment_ratios=None, disp=False):

    d = {'g': {}, 'bore_locations': coordinates, 'logtime': log_time}

    for i in range(len(H_values)):

        H = H_values[i]
        r_b = r_b_values[i]
        D = D_values[i]
        _borehole = gt.boreholes.Borehole(H, D, r_b, 0., 0.)

        alpha = soil.k / soil.rhoCp

        ts = H ** 2 / (9. * alpha)  # Bore field characteristic time
        time_values = np.exp(log_time) * ts

        gfunc = calculate_g_function(
            m_flow_borehole, bhe_object, time_values, coordinates, _borehole,
            fluid, pipe, grout, soil, nSegments=nSegments, segments=segments,
            solver=solver, boundary=boundary, segment_ratios=segment_ratios,
            disp=disp)

        key = '{}_{}_{}_{}'.format(B, H, r_b, D)

        d['g'][key] = gfunc.gFunc.tolist()

    geothermal_g_input = GFunction.configure_database_file_for_usage(d)
    # Initialize the GFunction object
    g_function = GFunction(**geothermal_g_input)

    return g_function


class GFunction:
    def __init__(self, B: float, r_b_values: dict, D_values: dict,
                 g_lts: dict, log_time: list, bore_locations: list):
        self.B: float = B  # a B spacing in the borefield
        # r_b (borehole radius) value keyed by height
        self.r_b_values: dict = r_b_values
        self.D_values: dict = D_values  # D (burial depth) value keyed by height
        self.g_lts: dict = g_lts  # g-functions (LTS) keyed by height
        # ln(t/ts) values that apply to all the heights
        self.log_time: list = log_time
        # (x, y) coordinates of boreholes
        self.bore_locations: list = bore_locations
        # self.time: dict = {}  # the time values in years

        # an interpolation table for B/H ratios, D, r_b (used in the method
        # g_function_interpolation)
        self.interpolation_table: dict = {}

    def g_function_interpolation(self, B_over_H: float, kind='default'):
        """
        Interpolate a range of g-functions for a specific B/H ratio
        Parameters
        ----------
        B_over_H: float
            A B/H ratio
        kind: str
            Could be 'linear', 'quadratic', 'cubic', etc.
            default: 'cubic'
        Returns
        -------
        **g-function: list**
            A list of the g-function values for each ln(t/ts)
        **rb: float**
            A borehole radius value that is interpolated for
        **D: float**
            A burial depth that is interpolated for
        **H_eq: float**
            An equivalent height
            .. math::
                H_{eq} = \dfrac{B_{field}}{B/H}
        """
        # the g-functions are stored in a dictionary based on heights, so an
        # equivalent height can be found
        H_eq = 1 / B_over_H * self.B

        # Determine if we are out of range and need to extrapolate
        height_values = list(self.g_lts.keys())
        # If we are close to the outer bounds, then set H_eq as outer bounds
        if abs(H_eq - max(height_values)) < 1.0e-6:
            H_eq = max(height_values)
        if abs(H_eq - min(height_values)) < 1.0e-6:
            H_eq = min(height_values)

        if min(height_values) <= H_eq <= max(height_values):
            fill_value = ''
        elif abs(min(height_values) - H_eq) < 0.001:
            fill_value = ''
        else:
            fill_value = 'extrapolate'
            warnings.warn('Extrapolation is being used.')

        # if the interpolation kind is default, use what we know about the
        # accuracy of interpolation to choose a technique
        if kind == 'default':
            num_curves = len(height_values)
            if num_curves >= 5:
                kind = 'cubic'
            elif num_curves >= 3:
                kind = 'quadratic'
            elif num_curves == 2:
                kind = 'linear'
            else:
                if (H_eq-height_values[0]) / height_values[0] < 0.001:
                    g_function = self.g_lts[height_values[0]]
                    rb = self.r_b_values[height_values[0]]
                    D = self.D_values[height_values[0]]
                    return g_function, rb, D, H_eq
                elif min(height_values) - H_eq < 0.001:
                    g_function = self.g_lts[height_values[0]]
                    rb = self.r_b_values[height_values[0]]
                    D = self.D_values[height_values[0]]
                    return g_function, rb, D, H_eq
                else:
                    raise ValueError(
                        'The interpolation requires two g-function curves if '
                        'the requested B/H is not already computed.')

        # Automatically adjust interpolation if necessary
        # Lagrange also needs 2
        interpolation_kinds = {'linear': 2, 'quadratic': 3, 'cubic': 4,
                               'lagrange': 2}
        curves_by_kind = {2: 'linear', 3: 'quadratic', 4: 'cubic'}
        if len(height_values) < 2:
            raise ValueError('Interpolation requires two g-function curves.')
        else:
            # If the number of required curves for the interpolation type is
            # greater than what is available, reduce the interpolation type
            required_curves = interpolation_kinds[kind]
            if required_curves > len(height_values):
                kind = curves_by_kind[len(height_values)]

        # if the interpolation table is not yet know, build it
        if len(self.interpolation_table) == 0:
            # create an interpolation for the g-function which takes the height
            # (or equivilant height) as an input the g-function needs to be
            # interpolated at each point in dimensionless time
            self.interpolation_table['g'] = []
            for i, lntts in enumerate(self.log_time):
                x = []
                y = []
                for key in self.g_lts:
                    height_value = float(key)
                    g_value = self.g_lts[key][i]
                    x.append(height_value)
                    y.append(g_value)
                if kind == 'lagrange':
                    f = lagrange(x, y)
                else:
                    f = interp1d(x, y, kind=kind, fill_value=fill_value)
                self.interpolation_table['g'].append(f)
            # create interpolation tables for 'D' and 'r_b' by height
            keys = list(self.r_b_values.keys())
            height_values: list = []
            rb_values: list = []
            D_values: list = []
            for h in keys:
                height_values.append(float(h))
                rb_values.append(self.r_b_values[h])
                try:
                    D_values.append(self.D_values[h])
                except:
                    pass
            if kind == 'lagrange':
                rb_f = lagrange(height_values, rb_values)
            else:
                # interpolation function for rb values by H equivalent
                rb_f = interp1d(height_values, rb_values, kind=kind,
                                fill_value=fill_value)
            self.interpolation_table['rb'] = rb_f
            try:
                if kind == 'lagrange':
                    D_f = lagrange(height_values, D_values)
                else:
                    D_f = interp1d(height_values, D_values, kind=kind,
                                   fill_value=fill_value)
                self.interpolation_table['D'] = D_f
            except:
                pass

        # create the g-function by interpolating at each ln(t/ts) value
        rb_value = self.interpolation_table['rb'](H_eq)
        try:
            D_value = self.interpolation_table['D'](H_eq)
        except:
            D_value = None
        g_function: list = []
        for i in range(len(self.log_time)):
            f = self.interpolation_table['g'][i]
            g = f(H_eq).tolist()
            g_function.append(g)
        return g_function, rb_value, D_value, H_eq

    @staticmethod
    def borehole_radius_correction(g_function: list, rb: float, rb_star: float):
        """
        Correct the borehole radius. From paper 3 of Eskilson 1987.
        .. math::
            g(\dfrac{t}{t_s}, \dfrac{r_b^*}{H}) =
            g(\dfrac{t}{t_s}, \dfrac{r_b}{H}) - ln(\dfrac{r_b^*}{r_b})
        Parameters
        ----------
        g_function: list
            A g-function
        rb: float
            The current borehole radius
        rb_star: float
            The borehole radius that is being corrected to
        Returns
        -------
        g_function_corrected: list
            A corrected g_function
        """
        g_function_corrected = []
        for i in range(len(g_function)):
            g_function_corrected.append(g_function[i] - np.log(rb_star / rb))
        return g_function_corrected

    def correct_coordinates(self, B: float):
        # scale the field by the ratio
        scale = B / self.B

        new_locations = []
        for i in range(len(self.bore_locations)):
            _x = self.bore_locations[i][0]
            _y = self.bore_locations[i][1]
            if _x == 0.0 and _y == 0.0:
                new_locations.append([_x, _y])
            elif _x == 0.0:
                new_locations.append([_x, _y*scale])
            else:
                x = _x * scale
                y = _y * scale
                new_locations.append([x, y])

        return new_locations

    def visualize_g_functions(self):
        """
        Visualize the g-functions.
        Returns
        -------
        **fig, ax**
            Figure and axes information.
        """
        fig = gt.utilities._initialize_figure()
        ax = fig.add_subplot(111)

        ax.set_xlim([-8.8, 3.9])
        ax.set_ylim([-2, 139])

        ax.text(2.75, 135, 'B/H')

        keys = reversed(list(self.g_lts.keys()))

        for key in keys:
            ax.plot(self.log_time, self.g_lts[key],
                    label=str(int(self.B)) + '/' + str(int(key)))
            x_n = self.log_time[-1]
            y_n = self.g_lts[key][-1]
            if key == 8:
                ax.annotate(str(round(float(self.B) / float(key), 4)),
                            xy=(x_n - .4, y_n - 5))
            else:
                ax.annotate(str(round(float(self.B) / float(key), 4)),
                            xy=(x_n-.4, y_n+1))

        handles, labels = ax.get_legend_handles_labels()

        legend = fig.legend(handles=handles, labels=labels,
                            title='B/H'.rjust(5), bbox_to_anchor=(1, 1.0))
        fig.gca().add_artist(legend)

        ax.set_ylabel('g-function')
        ax.set_xlabel('ln(t/t$_s$)')
        ax.grid()
        ax.set_axisbelow(True)
        fig.subplots_adjust(left=0.09, right=0.835, bottom=0.1, top=.99)

        return fig, ax

    def visualize_borefield(self, bore_locations=None):
        """
        Visualize the (x,y) coordinates.
        Returns
        -------
        **fig, ax**
            Figure and axes information.
        """
        if bore_locations is None:
            bore_locations = self.bore_locations
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(figsize=(3.5, 5))

        x, y = list(zip(*bore_locations))

        ax.scatter(x, y)

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')

        ax.set_aspect('equal')
        fig.tight_layout()

        return fig, ax

    @staticmethod
    def visualize_area_and_constraints(
            perimeter: list, coordinates: list, no_go: list = None):
        """
        Visualize the (x,y) coordinates, and no go zone.

        Returns
        -------
        **fig, ax**
            Figure and axes information.
        """
        fig = gt.utilities._initialize_figure()
        ax = fig.add_subplot(111)

        # Plot the perimeter if it's not None
        if len(perimeter) > 0:
            perimeter = perimeter + [perimeter[0]]
            _x, _y = list(zip(*perimeter))
            ax.plot(_x, _y, 'g')

        # Plot the coordinates if they are provided
        if len(coordinates) > 0:
            x, y = list(zip(*coordinates))
            ax.scatter(x, y)

        if no_go is not None:
            no_go = no_go + [no_go[0]]
            _x, _y = list(zip(*no_go))
            ax.plot(_x, _y, 'r')

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')

        fig.gca().set_aspect('equal')

        fig.tight_layout()

        return fig, ax

    @staticmethod
    def configure_database_file_for_usage(data) -> dict:
        """
        This method is called upon initialization of the object.
        Read the cpgfunction output dictionary into the borefield class for easy
        access of information
        Parameters
        ----------
        data: dict
            A dictionary which is in the output format of cpgfunction.
        Returns
        -------
            a dictionary for input to this object
        """
        log_time = data['logtime']

        bore_locations = data[
            'bore_locations']  # store the bore locations in the object
        g_values: dict = data['g']  # pull the g-functions into the g_values
        # a temporary g-function dictionary that might be out of order
        g_tmp: dict = {}
        # a temporary burial depth dictionary that may be out of order
        Ds_tmp: dict = {}
        # the borehole radius dictionary that may be out of order
        r_bs_tmp: dict = {}
        t_tmp: dict = {}
        for key in g_values:
            # do the g-function dictionary
            key_split = key.split('_')
            # get the current height
            height = float((key_split[1]))
            # create a g-function list associated with this height key
            g_tmp[height] = g_values[key]
            # create a r_b value associated with this height key
            r_b = float(key_split[2])
            r_bs_tmp[height] = r_b
            # the D value is recently added to the key value for the saved
            # g-functions computed
            try:
                D = float(key_split[3])
                Ds_tmp[height] = D
            except:
                pass

            # do the time dictionary
            time_arr: list = []
            for _, lntts in enumerate(log_time):
                alpha = 1.0e-06
                t_seconds = height ** 2 / 9 / alpha * np.exp(lntts)
                t_year = t_seconds / 60 / 24 / 365
                time_arr.append(t_year)
            t_tmp[height] = time_arr
        # every B-spacing should be the same for each file
        B = float(list(g_values.keys())[0].split('_')[0])

        keys = sorted(list(g_tmp.keys()), key=int)  # sort the heights in order
        # fill the g-function dictionary with sorted heights
        g = {key: g_tmp[key] for key in keys}
        # fill the burial depth dictionary with sorted heights
        try:
            Ds = {key: Ds_tmp[key] for key in keys}
        except:
            # if there's no D provided, make it 2
            Ds = {key: 2. for key in keys}
        r_bs = {key: r_bs_tmp[key] for key in keys}
        time = {key: t_tmp[key] for key in
                     keys}  # fill the time array for yearly points

        geothermal_g_input = {}
        geothermal_g_input['B'] = B
        geothermal_g_input['r_b_values'] = r_bs
        geothermal_g_input['D_values'] = Ds
        geothermal_g_input['g_lts'] = g
        geothermal_g_input['log_time'] = log_time
        geothermal_g_input['bore_locations'] = bore_locations

        return geothermal_g_input





def transpose_coordinates(coordinates):
    coordinates_transposed = []
    for i in range(len(coordinates)):
        x,y = coordinates[i]
        coordinates_transposed.append((y, x))
    return coordinates_transposed


def rectangle(Nx, Ny, Bx, By, origin=(0, 0)):
    # Create a list of (x, y) pairs for a rectangle
    r = []
    nbh = Nx * Ny
    _x = origin[0]
    _y = origin[1]
    for i in range(Nx):
        for j in range(Ny):
            r.append((_x + i * Bx, _y + j * By))
    assert len(r) == nbh
    return r


def open_rectangle(Nx, Ny, Bx, By):
    # Create a list of (x, y) pairs for an open rectangle
    open_r = []
    if Nx > 2 and Ny > 2:
        nbh = Ny * 2 + (Nx - 2) * 2
        for i in range(Nx):
            open_r.append((i * Bx, 0.))
        for j in range(1, Ny-1):
            open_r.append((0, j * By))
            open_r.append(((Nx-1) * Bx, j * By))
        for i in range(Nx):
            open_r.append((i * Bx, (Ny-1) * By))
    else:
        nbh = Nx * Ny
        open_r = rectangle(Nx, Ny, Bx, By)

    assert len(open_r) == nbh
    return open_r


def C_shape(Nx_1, Ny, Bx, By, Nx_2):
    nbh = Nx_1 + (Ny * 2) - 1 + Nx_2 - 1
    c = []
    for i in range(Nx_1):
        c.append((i * Bx, 0.))
    x_loc = (Nx_1 - 1) * Bx
    for j in range(1, Ny):
        c.append((0., j * By))
    for j in range(1, Ny):
        c.append((x_loc, j * By))
    y_loc = (Ny-1) * By
    for i in range(1, Nx_2+1):
        c.append((i * Bx, y_loc))
    assert len(c) == nbh
    return c


def U_shape(Nx, Ny, Bx, By):
    # Create a list of (x, y) pairs for a U-shape
    U = []
    if Nx > 2 and Ny > 1:
        nbh = 2 * Ny + (Nx - 2)
        for i in range(Nx):
            U.append((i * Bx, 0.))
        for j in range(1, Ny):
            U.append((0., j * By))
    else:
        nbh = Nx * Ny
        U = rectangle(Nx, Ny, Bx, By)
    assert len(U) == nbh
    return U


def lop_U(Nx, Ny_1, Bx, By, Ny_2):
    nbh = Nx + Ny_1 - 1 + Ny_2 - 1
    lop_u = []
    for i in range(Nx):
        lop_u.append((i * Bx, 0.))
    for j in range(1, Ny_1):
        lop_u.append((0., j * By))
    x_loc = (Nx-1)*Bx
    for j in range(1, Ny_2):
        lop_u.append((x_loc, j * By))
    assert len(lop_u) == nbh
    return lop_u


def L_shape(Nx, Ny, Bx, By):
    nbh = Nx + Ny - 1
    L = []
    for i in range(Nx):
        L.append((i * Bx, 0.))
    for j in range(1, Ny):
        L.append((0., j * By))
    assert len(L) == nbh
    return L


def zoned_rectangle(Nx, Ny, Bx, By, Nix, Niy):
    # Create a zoned rectangle
    # The creator of the idea behind the "zoned rectangle" is
    # Dr. Jeffrey D. Spitler

    if Nix > (Nx - 2):
        raise ValueError('To many interior x boreholes.')
    if Niy > (Ny - 2):
        raise ValueError('Too many interior y boreholes.')

    # Create a list of (x, y) coordinates
    zoned = []

    # Boreholes on the perimeter
    zoned.extend(open_rectangle(Nx, Ny, Bx, By))

    # Create the interior coordinates
    Bix = (Nx - 1) * Bx / (Nix + 1)
    Biy = (Ny - 1) * By / (Niy + 1)

    zoned.extend(rectangle(Nix, Niy, Bix, Biy, origin=(Bix, Biy)))

    return zoned


def visualize_coordinates(coordinates):
    """
    Visualize the (x,y) coordinates.
    Returns
    -------
    **fig, ax**
        Figure and axes information.
    """
    import matplotlib.pyplot as plt
    plt.rc('font', size=9)
    plt.rc('xtick', labelsize=9)
    plt.rc('ytick', labelsize=9)
    plt.rc('lines', lw=1.5, markersize=5.0)
    plt.rc('savefig', dpi=500)
    # fig, ax = plt.subplots(figsize=(3.5, 5))
    import pygfunction as gt
    fig = gt.utilities._initialize_figure()
    ax = fig.add_subplot(111)

    x, y = list(zip(*coordinates))

    ax.scatter(x, y)

    ax.set_xlabel('x (m)')
    ax.set_ylabel('y (m)')

    ax.set_aspect('equal')

    return fig, ax



import ghedt as dt
import pygfunction as gt
import ghedt.peak_load_analysis_tool as plat
from ghedt.utilities import sign, check_bracket
import numpy as np
import copy
from sys import stderr


class Bisection1D:
    def __init__(self, coordinates_domain: list, V_flow: float,
                 borehole: gt.boreholes.Borehole,
                 bhe_object: plat.borehole_heat_exchangers,
                 fluid: gt.media.Fluid, pipe: plat.media.Pipe,
                 grout: plat.media.Grout, soil: plat.media.Soil,
                 sim_params: plat.media.SimulationParameters,
                 hourly_extraction_ground_loads: list, method: str = 'hybrid',
                 flow: str = 'borehole', max_iter=15, disp=False, search=True):

        # Take the lowest part of the coordinates domain to be used for the
        # initial setup
        coordinates = coordinates_domain[0]

        # Flow rate tracking
        self.V_flow = V_flow
        self.flow = flow
        V_flow_system, m_flow_borehole = \
            self.retrieve_flow(coordinates, fluid.rho)
        self.method = method

        self.log_time = dt.utilities.Eskilson_log_times()
        self.bhe_object = bhe_object
        self.sim_params = sim_params
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads
        self.coordinates_domain = coordinates_domain
        self.max_iter = max_iter
        self.disp = disp

        B = dt.utilities.borehole_spacing(borehole, coordinates)

        # Calculate a g-function for uniform inlet fluid temperature with
        # 8 unequal segments using the equivalent solver
        g_function = dt.gfunction.compute_live_g_function(
            B, [borehole.H], [borehole.r_b], [borehole.D], m_flow_borehole,
            self.bhe_object, self.log_time, coordinates, fluid, pipe, grout,
            soil)

        # Initialize the GHE object
        self.ghe = dt.ground_heat_exchangers.GHE(
            V_flow_system, B, bhe_object, fluid, borehole, pipe, grout,
            soil, g_function, sim_params, hourly_extraction_ground_loads)

        self.calculated_temperatures = {}

        if search:
            self.selection_key, self.selected_coordinates = self.search()

    def retrieve_flow(self, coordinates, rho):
        if self.flow == 'borehole':
            V_flow_system = self.V_flow * float(len(coordinates))
            # Total fluid mass flow rate per borehole (kg/s)
            m_flow_borehole = self.V_flow / 1000. * rho
        elif self.flow == 'system':
            V_flow_system = self.V_flow
            V_flow_borehole = self.V_flow / float(len(coordinates))
            m_flow_borehole = V_flow_borehole / 1000. * rho
        else:
            raise ValueError('The flow argument should be either `borehole`'
                             'or `system`.')
        return V_flow_system, m_flow_borehole

    def initialize_ghe(self, coordinates, H):
        V_flow_system, m_flow_borehole = \
            self.retrieve_flow(coordinates, self.ghe.bhe.fluid.rho)

        self.ghe.bhe.b.H = H
        borehole = self.ghe.bhe.b
        fluid = self.ghe.bhe.fluid
        pipe = self.ghe.bhe.pipe
        grout = self.ghe.bhe.grout
        soil = self.ghe.bhe.soil

        B = dt.utilities.borehole_spacing(borehole, coordinates)

        # Calculate a g-function for uniform inlet fluid temperature with
        # 8 unequal segments using the equivalent solver
        g_function = dt.gfunction.compute_live_g_function(
            B, [borehole.H], [borehole.r_b], [borehole.D], m_flow_borehole,
            self.bhe_object, self.log_time, coordinates, fluid, pipe, grout,
            soil)

        # Initialize the GHE object
        self.ghe = dt.ground_heat_exchangers.GHE(
            V_flow_system, B, self.bhe_object, fluid, borehole, pipe, grout,
            soil, g_function, self.sim_params,
            self.hourly_extraction_ground_loads)

    def calculate_excess(self, coordinates, H):
        self.initialize_ghe(coordinates, H)
        # Simulate after computing just one g-function
        max_HP_EFT, min_HP_EFT = self.ghe.simulate(method=self.method)
        T_excess = self.ghe.cost(max_HP_EFT, min_HP_EFT)

        # This is more of a debugging statement. May remove it in the future.
        # Perhaps there becomes a debug: bool option in the API.
        # if self.disp:
        #     print('Min EFT: {}\nMax EFT: {}'.format(min_HP_EFT, max_HP_EFT))

        return T_excess

    def search(self):

        xL_idx = 0
        xR_idx = len(self.coordinates_domain) - 1
        if self.disp:
            print('Do some initial checks before searching.')
        _nbh_max_warn = 2500
        if len(self.coordinates_domain[xR_idx]) > _nbh_max_warn:
            print(f'Info: The maximum number of boreholes in the domain '
                  f'exceeds {_nbh_max_warn} boreholes. Expect the initial '
                  f'checks to take some time. To decrease search time, '
                  f'consider increasing the minimum borehole spacing.',
                  file=stderr)
        # Get the lowest possible excess temperature from minimum height at the
        # smallest location in the domain
        T_0_lower = self.calculate_excess(self.coordinates_domain[xL_idx],
                                          self.sim_params.min_Height)
        T_0_upper = self.calculate_excess(self.coordinates_domain[xL_idx],
                                          self.sim_params.max_Height)
        T_m1 = \
            self.calculate_excess(
                self.coordinates_domain[xR_idx],
                self.sim_params.max_Height)

        self.calculated_temperatures[xL_idx] = T_0_upper
        self.calculated_temperatures[xR_idx] = T_m1

        if check_bracket(sign(T_0_lower), sign(T_0_upper)):
            if self.disp:
                print('Size between min and max of lower bound in domain.')
            self.initialize_ghe(self.coordinates_domain[0],
                                self.sim_params.max_Height)
            return 0, self.coordinates_domain[0]
        elif check_bracket(sign(T_0_upper), sign(T_m1)):
            if self.disp:
                print('Perform the integer bisection search routine.')
            pass
        else:
            # This domain does not bracked the solution
            if T_0_upper < 0.0 and T_m1 < 0.0:
                msg = 'Based on the loads provided, the excess temperatures ' \
                      'for the minimum and maximum number of boreholes falls ' \
                      'below 0. This means that the loads are "miniscule" or ' \
                      'that the lower end of the domain needs to contain ' \
                      'less boreholes.'.format()
                raise ValueError(msg)
            if T_0_upper > 0.0 and T_m1 > 0.0:
                msg = 'Based on the loads provided, the excess temperatures ' \
                      'for the minimum and maximum number of boreholes falls ' \
                      'above 0. This means that the loads are "astronomical" ' \
                      'or that the higher end of the domain needs to contain ' \
                      'more boreholes. Consider increasing the available land' \
                      ' area, or decreasing the minimum allowable borehole ' \
                      'spacing.'
                raise ValueError(msg)
            return None, None

        if self.disp:
            print('Beginning bisection search...')

        xL_sign = sign(T_0_upper)
        xR_sign = sign(T_m1)
        
        i = 0

        while i < self.max_iter:
            c_idx = int(np.ceil((xL_idx + xR_idx) / 2))
            # if the solution is no longer making progress break the while
            if c_idx == xL_idx or c_idx == xR_idx:
                break

            c_T_excess = self.calculate_excess(self.coordinates_domain[c_idx],
                                               self.sim_params.max_Height)
            self.calculated_temperatures[c_idx] = c_T_excess
            c_sign = sign(c_T_excess)

            if c_sign == xL_sign:
                xL_idx = copy.deepcopy(c_idx)
            else:
                xR_idx = copy.deepcopy(c_idx)

            i += 1

        coordinates = self.coordinates_domain[i]

        H = self.sim_params.max_Height

        self.calculate_excess(coordinates, H)
        # Make sure the field being returned pertains to the index which is the
        # closest to 0 but also negative (the maximum of all 0 or negative
        # excess temperatures)
        keys = list(self.calculated_temperatures.keys())
        values = list(self.calculated_temperatures.values())

        negative_excess_values = [values[i] for i in range(len(values))
                                  if values[i] <= 0.0]

        excess_of_interest = max(negative_excess_values)
        idx = values.index(excess_of_interest)
        selection_key = keys[idx]
        selected_coordinates = self.coordinates_domain[selection_key]

        self.initialize_ghe(selected_coordinates, H)

        return selection_key, selected_coordinates


class Bisection2D(Bisection1D):
    def __init__(self, coordinates_domain_nested: list, V_flow: float,
                 borehole: gt.boreholes.Borehole,
                 bhe_object: plat.borehole_heat_exchangers,
                 fluid: gt.media.Fluid, pipe: plat.media.Pipe,
                 grout: plat.media.Grout, soil: plat.media.Soil,
                 sim_params: plat.media.SimulationParameters,
                 hourly_extraction_ground_loads: list, method: str = 'hybrid',
                 flow: str = 'borehole', max_iter=15, disp=False):
        if disp:
            print('Note: This routine requires a nested bisection search.')

        # Get a coordinates domain for initialization
        coordinates_domain = coordinates_domain_nested[0]
        Bisection1D.__init__(
            self, coordinates_domain, V_flow, borehole, bhe_object,
            fluid, pipe, grout, soil, sim_params,
            hourly_extraction_ground_loads, method=method, flow=flow,
            max_iter=max_iter, disp=disp, search=False)

        self.coordinates_domain_nested = []
        self.calculated_temperatures_nested = []
        # Tack on one borehole at the beginning to provide a high excess
        # temperature
        outer_domain = [coordinates_domain_nested[0][0]]
        for i in range(len(coordinates_domain_nested)):
            outer_domain.append(coordinates_domain_nested[i][-1])

        self.coordinates_domain = outer_domain

        selection_key, selected_coordinates = self.search()

        self.calculated_temperatures_nested.append(
            copy.deepcopy(self.calculated_temperatures))

        # We tacked on one borehole to the beginning, so we need to subtract 1
        # on the index
        inner_domain = coordinates_domain_nested[selection_key-1]
        self.coordinates_domain = inner_domain

        # Reset calculated temperatures
        self.calculated_temperatures = {}

        self.selection_key, self.selected_coordinates = self.search()


class BisectionZD(Bisection1D):
    def __init__(self, coordinates_domain_nested: list, V_flow: float,
                 borehole: gt.boreholes.Borehole,
                 bhe_object: plat.borehole_heat_exchangers,
                 fluid: gt.media.Fluid, pipe: plat.media.Pipe,
                 grout: plat.media.Grout, soil: plat.media.Soil,
                 sim_params: plat.media.SimulationParameters,
                 hourly_extraction_ground_loads: list, method: str = 'hybrid',
                 flow: str = 'borehole', max_iter=15, disp=False):
        if disp:
            print('Note: This design routine currently requires several '
                  'bisection searches.')

        # Get a coordinates domain for initialization
        coordinates_domain = coordinates_domain_nested[0]
        Bisection1D.__init__(
            self, coordinates_domain, V_flow, borehole, bhe_object,
            fluid, pipe, grout, soil, sim_params,
            hourly_extraction_ground_loads, method=method, flow=flow,
            max_iter=max_iter, disp=disp, search=False)

        self.coordinates_domain_nested = coordinates_domain_nested
        self.calculated_temperatures_nested = {}
        # Tack on one borehole at the beginning to provide a high excess
        # temperature
        outer_domain = [coordinates_domain_nested[0][0]]
        for i in range(len(coordinates_domain_nested)):
            outer_domain.append(coordinates_domain_nested[i][-1])

        self.coordinates_domain = outer_domain

        self.selection_key_outer, self.selected_coordinates_outer = \
            self.search()

        if self.selection_key_outer > 0:
            self.selection_key_outer -= 1
        self.calculated_heights = {}

        self.selection_key, self.selected_coordinates = self.search_successive()

    def search_successive(self, max_iter=None):
        if max_iter is None:
            max_iter = self.selection_key_outer + 7

        i = self.selection_key_outer

        old_height = 99999

        while i < len(self.coordinates_domain_nested) and i < max_iter:

            self.coordinates_domain = self.coordinates_domain_nested[i]
            self.calculated_temperatures = {}
            try:
                selection_key, selected_coordinates = self.search()
            except ValueError:
                break
            self.calculated_temperatures_nested[i] = \
                copy.deepcopy(self.calculated_temperatures)

            self.ghe.compute_g_functions()
            self.ghe.size(method='hybrid')

            nbh = len(selected_coordinates)
            total_drilling = float(nbh) * self.ghe.bhe.b.H
            self.calculated_heights[i] = total_drilling

            if old_height < total_drilling:
                break
            else:
                old_height = copy.deepcopy(total_drilling)

            i += 1

        keys = list(self.calculated_heights.keys())
        values = list(self.calculated_heights.values())

        minimum_total_drilling = min(values)
        idx = values.index(minimum_total_drilling)
        selection_key_outer = keys[idx]
        self.calculated_temperatures = \
            copy.deepcopy(self.calculated_temperatures_nested[
                              selection_key_outer])

        keys = list(self.calculated_temperatures.keys())
        values = list(self.calculated_temperatures.values())

        negative_excess_values = [values[i] for i in range(len(values))
                                  if values[i] <= 0.0]

        excess_of_interest = max(negative_excess_values)
        idx = values.index(excess_of_interest)
        selection_key = keys[idx]
        selected_coordinates = \
            self.coordinates_domain_nested[selection_key_outer][selection_key]

        self.initialize_ghe(selected_coordinates, self.sim_params.max_Height)
        self.ghe.compute_g_functions()
        self.ghe.size(method='hybrid')

        return selection_key, selected_coordinates


# The following functions are utility functions specific to search_routines.py
# ------------------------------------------------------------------------------
def oak_ridge_export(bisection_search, file_name='ghedt_output'):
    # Dictionary for export
    d = {}
    d['borehole_length'] = bisection_search.ghe.bhe.b.H
    d['number_of_boreholes'] = len(bisection_search.selected_coordinates)
    d['g_function_pairs'] = []
    d['single_u_tube'] = {}

    # create a local single U-tube object
    bhe_eq = bisection_search.ghe.bhe_eq
    d['single_u_tube']['r_b'] = bhe_eq.borehole.r_b  # Borehole radius
    d['single_u_tube']['r_in'] = bhe_eq.r_in  # Inner pipe radius
    d['single_u_tube']['r_out'] = bhe_eq.r_out  # Outer pipe radius
    # Note: Shank spacing or center pipe positions could be used
    d['single_u_tube']['s'] = bhe_eq.pipe.s  # Shank spacing (tube-to-tube)
    d['single_u_tube']['pos'] = bhe_eq.pos  # Center of the pipes
    d['single_u_tube']['m_flow_borehole'] = \
        bhe_eq.m_flow_borehole  # mass flow rate of the borehole
    d['single_u_tube']['k_g'] = bhe_eq.grout.k  # Grout thermal conductivity
    d['single_u_tube']['k_s'] = bhe_eq.soil.k  # Soil thermal conductivity
    d['single_u_tube']['k_p'] = bhe_eq.pipe.k  # Pipe thermal conductivity

    # create a local ghe object
    ghe = bisection_search.ghe
    H = ghe.bhe.b.H
    B_over_H = ghe.B_spacing / H
    g = ghe.grab_g_function(B_over_H)

    total_g_values = g.x.size
    number_lts_g_values = 27
    number_sts_g_values = 30
    sts_step_size = int(np.floor((total_g_values - number_lts_g_values) /
                                 number_sts_g_values).tolist())
    lntts = []
    g_values = []
    for i in range(1, (total_g_values - number_lts_g_values),
                   sts_step_size):
        lntts.append(g.x[i].tolist())
        g_values.append(g.y[i].tolist())
    lntts += g.x[
             total_g_values - number_lts_g_values: total_g_values].tolist()
    g_values += g.y[
                total_g_values - number_lts_g_values: total_g_values].tolist()

    for i in range(len(lntts)):
        d['g_function_pairs'].append({'ln_tts': lntts[i],
                                      'g_value': g_values[i]})

    dt.utilities.js_dump(file_name, d, indent=4)

    return





class GeometricConstraints:
    def __init__(self, length: float = None, width: float = None,
                 B: float = None, B_min: float = None,
                 B_max_x: float = None, B_max_y: float = None,
                 outer_constraints: list = None, no_go: list = None):
        # Spacing parameters in meters
        self.B = B
        self.B_max_x = B_max_x
        self.B_max_y = B_max_y
        self.B_min = B_min
        # Length and width constraints
        self.length = length
        self.width = width
        # Outer constraints described as a polygon
        self.outer_constraints = outer_constraints
        # TODO: Handling for a list or a list of lists to occur later
        # Note: the entirety of the no-go zone should fall inside the
        # outer_constraints
        self.no_go = no_go

    def check_inputs(self, method):
        # The required instances for the near-square design is self.B
        if method == 'near-square':
            assert self.B is not None
            assert self.length is not None
        elif method == 'rectangle':
            assert self.width is not None
            assert self.length is not None
            assert self.B_min is not None
            assert self.B_max_x is not None
        elif method == 'bi-rectangle' or 'bi-zoned':
            assert self.width is not None
            assert self.length is not None
            assert self.B_min is not None
            assert self.B_max_x is not None
            assert self.B_max_y is not None

        return



import pygfunction as gt
import ghedt.peak_load_analysis_tool as plat
import numpy as np
from numpy import pi


class BasePipe(object):
    def __init__(self, m_flow_borehole, fluid, borehole, soil, grout, pipe,
                 config=None):
        # borehole mass flow rate
        self.m_flow_borehole = m_flow_borehole
        # mass flow rate to each pipe
        self.m_flow_pipe = self.compute_mass_flow_rate_pipe(m_flow_borehole,
                                                            config)
        # Store Thermal properties
        self.soil = soil
        self.grout = grout
        self.pipe = pipe
        # Store fluid properties
        self.fluid = fluid
        # Store borehole
        self.borehole = borehole

        # Declare variables that are computed in compute_resistance()
        self.h_f = None
        self.R_p = None
        self.R_f = None

        self.compute_resistances()

    @staticmethod
    def justify(category, value):
        return category.ljust(40) + '= ' + value + '\n'

    def __repr__(self):
        justify = self.justify

        output = str(self.__class__) + '\n'
        output += 50 * '-' + '\n'

        output += justify('Mass flow borehole',
                          str(round(self.m_flow_borehole, 4)) + ' (kg/s)')
        output += justify('Mass flow pipe',
                          str(round(self.m_flow_pipe, 4)) + ' (kg/s)')
        output += self.borehole.__repr__() + '\n'
        output += self.soil.__repr__()
        output += self.grout.__repr__()
        output += self.pipe.__repr__()
        output += self.fluid.__repr__()
        output += '\n'

        Re = compute_Reynolds_concentric(
            self.m_flow_pipe, self.pipe.r_in, self.pipe.eps, self.fluid)

        output += justify('Reynolds number', str(round(Re, 4)))
        output += justify('Convection coefficient',
                          str(round(self.h_f, 4)) + ' (W/m2.K)')
        output += justify('Pipe resistance',
                          str(round(self.R_p, 4)) + ' (m.K/W)')
        output += justify('Fluid resistance',
                          str(round(self.R_f, 4)) + ' (m.K/W)')

        return output

    def compute_resistances(self):
        # Convection heat transfer coefficient of a single pipe
        self.h_f = \
            gt.pipes.convective_heat_transfer_coefficient_circular_pipe(
                self.m_flow_pipe, self.pipe.r_in, self.fluid.mu, self.fluid.rho,
                self.fluid.k, self.fluid.cp, self.pipe.eps)

        # Single pipe thermal resistance
        self.R_p = gt.pipes.conduction_thermal_resistance_circular_pipe(
            self.pipe.r_in, self.pipe.r_out, self.pipe.k)
        # Single pipe fluid thermal resistance
        self.R_f = 1.0 / (self.h_f * 2 * pi * self.pipe.r_in)

        return

    @staticmethod
    def compute_mass_flow_rate_pipe(m_flow_borehole, config):
        if config == 'series' or config is None:
            m_flow_pipe = m_flow_borehole
        elif config == 'parallel':
            m_flow_pipe = m_flow_borehole / 2.
        else:
            raise ValueError('No such configuration exists.')
        return m_flow_pipe

    def compute_effective_borehole_resistance(
            self, m_flow_borehole=None, fluid=None):
        # Compute the effective borehole thermal resistance

        # if the mass flow rate has changed, then update it and use new value
        if m_flow_borehole is None:
            m_flow_borehole = self.m_flow_borehole
        else:
            self.m_flow_borehole = m_flow_borehole

        # if the mass flow rate has changed, then update it and use new value
        if fluid is None:
            fluid = self.fluid
        else:
            self.fluid = fluid

        R_b_star = self.effective_borehole_thermal_resistance(
            m_flow_borehole, fluid.cp)

        return R_b_star


class SingleUTube(BasePipe, gt.pipes.SingleUTube):
    def __init__(self, m_flow_borehole: float,
                 fluid: gt.media.Fluid, borehole: gt.boreholes.Borehole,
                 pipe: plat.media.Pipe, grout: plat.media.Grout,
                 soil: plat.media.Soil):
        # Initialize base pipe class
        BasePipe.__init__(
            self, m_flow_borehole, fluid, borehole, soil, grout, pipe)
        # Compute Single pipe and fluid thermal resistance
        R_fp = self.R_f + self.R_p
        # Initialize pygfunction SingleUTube base class
        gt.pipes.SingleUTube.__init__(
            self, pipe.pos, pipe.r_in, pipe.r_out, borehole, self.soil.k,
            self.grout.k, R_fp)

    def __repr__(self):
        justify = self.justify

        output = BasePipe.__repr__(self)

        Rb_star = self.effective_borehole_thermal_resistance(
            self.m_flow_borehole, self.fluid.cp)

        output += justify('Effective borehole resistance',
                          str(round(Rb_star, 4)) + ' (m.K/W)')

        return output

    def update_thermal_resistance(self, m_flow_borehole=None, fluid=None):

        # if the mass flow rate has changed, then update it and use new value
        if m_flow_borehole is None:
            m_flow_borehole = self.m_flow_borehole
        else:
            self.m_flow_borehole = m_flow_borehole

        # if the mass flow rate has changed, then update it and use new value
        if fluid is None:
            fluid = self.fluid
        else:
            self.fluid = fluid

        self.compute_resistances()
        R_fp = self.R_f + self.R_p
        self.update_thermal_resistances(R_fp)

        # Note: This is the local borehole resistance
        # # Evaluate borehole thermal resistance
        # R_b = 1 / np.trace(1 / self._Rd)

        # Compute and return effective borehole resistance
        R_b_star = self.effective_borehole_thermal_resistance(
            m_flow_borehole, fluid.cp)

        return R_b_star

    def compute_convection_coefficient(self, m_flow_borehole):
        h = gt.pipes.convective_heat_transfer_coefficient_circular_pipe(
            m_flow_borehole, self.pipe.r_in, self.fluid.mu, self.fluid.rho,
            self.fluid.k, self.fluid.cp, self.pipe.eps)
        return h


class MultipleUTube(BasePipe, gt.pipes.MultipleUTube):
    def __init__(self, m_flow_borehole: float, fluid: gt.media.Fluid,
                 borehole: gt.boreholes.Borehole,
                 pipe: plat.media.Pipe, grout: plat.media.Grout,
                 soil: plat.media.Soil, config='parallel'):
        # Initialize base pipe class
        BasePipe.__init__(
            self, m_flow_borehole, fluid, borehole, soil, grout, pipe,
            config=config)

        # Get number of pipes from positions
        self.n_pipes = int(len(pipe.pos) / 2)
        # Compute Single pipe and fluid thermal resistance
        R_fp = self.R_f + self.R_p
        gt.pipes.MultipleUTube.__init__(
            self, pipe.pos, pipe.r_in, pipe.r_out, borehole, self.soil.k,
            self.grout.k, R_fp, pipe.n_pipes, config=config)

    def update_thermal_resistance(self, m_flow_borehole=None, fluid=None):

        # if the mass flow rate has changed, then update it and use new value
        if m_flow_borehole is None:
            m_flow_borehole = self.m_flow_borehole
        else:
            self.m_flow_borehole = m_flow_borehole

        # if the mass flow rate has changed, then update it and use new value
        if fluid is None:
            fluid = self.fluid
        else:
            self.fluid = fluid

        self.compute_resistances()

        self.compute_resistances()
        R_fp = self.R_f + self.R_p
        self.update_thermal_resistances(R_fp)

        # Note: This is the local borehole resistance
        # # Evaluate borehole thermal resistance
        # R_b = 1 / np.trace(1 / self._Rd)

        # Compute and return effective borehole resistance
        R_b_star = self.effective_borehole_thermal_resistance(
            m_flow_borehole, fluid.cp)

        return R_b_star

    def __repr__(self):
        justify = self.justify

        output = BasePipe.__repr__(self)

        Rb_star = self.effective_borehole_thermal_resistance(
            self.m_flow_borehole, self.fluid.cp)

        output += justify('Effective borehole resistance',
                          str(round(Rb_star, 4)) + ' (m.K/W)')

        output += justify('Config', self.config)

        return output


class CoaxialBase(object):
    def __init__(self, m_flow_borehole, fluid, borehole, pipe, soil, grout,
                 config=None):
        self.m_flow_borehole = m_flow_borehole
        self.m_flow_pipe = m_flow_borehole
        # Store Thermal properties
        self.soil = soil
        self.grout = grout
        self.pipe = pipe
        # Store fluid properties
        self.fluid = fluid
        # Store pipe roughness
        self.eps = self.pipe.eps

        self.r_inner = pipe.r_in
        self.r_outer = pipe.r_out

        # Unpack the radii to reduce confusion in the future
        self.r_in_in, self.r_in_out = self.r_inner
        self.r_out_in, self.r_out_out = self.r_outer

        self.b = borehole  # pygfunction borehole

        # Declare variables that are computed in compute_resistance()
        self.R_p_in = None
        self.R_p_out = None
        self.R_grout = None
        self.h_fluid_a_in = None
        self.h_fluid_a_out = None
        self.R_f_a_in = None
        self.R_f_a_out = None
        self.h_fluid_in = None
        self.R_f_in = None
        self.R_ff = None
        self.R_fp = None

        self.compute_resistances()

        return

    @staticmethod
    def justify(category, value):
        return category.ljust(40) + '= ' + value + '\n'

    def __repr__(self):
        justify = self.justify

        output = str(self.__class__) + '\n'
        output += 50 * '-' + '\n'

        output += justify('Mass flow borehole',
                          str(round(self.m_flow_borehole, 4)) + ' (kg/s)')
        output += justify('Mass flow pipe',
                          str(round(self.m_flow_pipe, 4)) + ' (kg/s)')
        output += self.b.__repr__() + '\n'
        output += self.soil.__repr__()
        output += self.grout.__repr__()
        output += self.pipe.__repr__()
        output += self.fluid.__repr__()
        output += '\n'

        output += justify('Inner pipe resistance',
                          str(round(self.R_p_in, 4)) + ' (m.K/W)')
        output += justify('Outer pipe resistance',
                          str(round(self.R_p_out, 4)) + ' (m.K/W)')
        output += justify('Grout resistance',
                          str(round(self.R_grout, 4)) + ' (m.K/W)')
        Re_annulus = compute_Reynolds_concentric(
            self.m_flow_pipe, self.r_in_out, self.r_out_in, self.fluid
        )
        output += justify('Reynolds annulus', str(round(Re_annulus, 4)))
        output += justify('Inner annulus convection coef',
                          str(round(self.h_fluid_a_in, 4)) + ' (W/m2.K)')
        output += justify('Outer annulus convection coef',
                          str(round(self.h_fluid_a_out, 4)) + ' (W/m2.K)')
        output += justify('Inner annulus resistance',
                          str(round(self.R_f_a_in, 4)) + ' (m.K/W)')
        output += justify('Outer annulus resistance',
                          str(round(self.R_f_a_out, 4)) + ' (m.K/W)')
        Re = compute_Reynolds(
            self.m_flow_pipe, self.r_in_in, self.eps, self.fluid)
        output += justify('Reynolds inner', str(round(Re, 4)))
        output += justify('Inner Convection coefficient',
                          str(round(self.h_fluid_in, 4)) + ' (W/m2.K)')
        output += justify('Fluid-to-fluid resistance',
                          str(round(self.R_ff, 4)) + ' (m.K/W)')
        output += justify('Fluid-to-pipe resistance',
                          str(round(self.R_fp, 4)) + ' (m.K/W)')

        Rb_star = self.effective_borehole_thermal_resistance(
            self.m_flow_borehole, self.fluid.cp)

        output += justify('Effective borehole resistance',
                          str(round(Rb_star, 4)) + ' (m.K/W)')

        return output

    def compute_resistances(self):
        # Inner pipe thermal resistance
        self.R_p_in = gt.pipes. \
            conduction_thermal_resistance_circular_pipe(self.r_in_in,
                                                        self.r_in_out,
                                                        self.pipe.k[0])
        # Outer pipe thermal resistance
        self.R_p_out = gt.pipes. \
            conduction_thermal_resistance_circular_pipe(self.r_out_in,
                                                        self.r_out_out,
                                                        self.pipe.k[1])
        # Grout thermal resistance
        self.R_grout = gt.pipes.conduction_thermal_resistance_circular_pipe(
            self.r_out_out, self.b.r_b, self.grout.k)

        h_fluid_a_in, h_fluid_a_out = \
            gt.pipes.convective_heat_transfer_coefficient_concentric_annulus(
                self.m_flow_borehole, self.r_in_out, self.r_out_in,
                self.fluid.mu, self.fluid.rho, self.fluid.k, self.fluid.cp,
                self.eps)

        self.h_fluid_a_in = h_fluid_a_in
        self.h_fluid_a_out = h_fluid_a_out

        # Inner fluid convective resistance
        self.R_f_a_in = 1. / (self.h_fluid_a_in * 2 * pi * self.r_in_out)
        # Outer fluid convective resistance
        self.R_f_a_out = 1. / (h_fluid_a_out * 2 * pi * self.r_out_in)
        self.h_fluid_in = \
            gt.pipes.convective_heat_transfer_coefficient_circular_pipe(
                self.m_flow_borehole, self.r_in_in, self.fluid.mu,
                self.fluid.rho, self.fluid.k, self.fluid.cp, self.eps)
        self.R_f_in = 1. / (self.h_fluid_in * 2 * pi * self.r_in_in)

        return


class CoaxialPipe(CoaxialBase, gt.pipes.Coaxial, BasePipe):
    def __init__(self, m_flow_borehole: float, fluid: gt.media.Fluid,
                 borehole: gt.boreholes.Borehole,
                 pipe: plat.media.Pipe, grout: plat.media.Grout,
                 soil: plat.media.Soil, config=None):
        CoaxialBase.__init__(self, m_flow_borehole, fluid, borehole, pipe,
                             soil, grout, config=config)

        # Vectors of inner and outer pipe radii
        # Note : The dimensions of the inlet pipe are the first elements of
        #       the vectors. In this example, the inlet pipe is the inside pipe.
        r_inner_p = np.array([pipe.r_in[0], pipe.r_out[0]])  # Inner pipe radii (m)
        r_outer_p = np.array([pipe.r_in[1], pipe.r_out[1]])  # Outer pip radii (m)

        # Compute resistances that are stored by pygfunction
        # Inner fluid to inner anulus fluid resistance
        R_ff = self.R_f_in + self.R_p_in + self.R_f_a_in
        # Outer annulus fluid to pipe thermal resistance
        R_fp = self.R_p_out + self.R_f_a_out

        gt.pipes.Coaxial.__init__(
            self, pipe.pos, r_inner_p, r_outer_p,  borehole, self.soil.k,
            self.grout.k, R_ff, R_fp, J=2)

    def update_thermal_resistance(self, m_flow_borehole=None, fluid=None):

        # if the mass flow rate has changed, then update it and use new value
        if m_flow_borehole is None:
            m_flow_borehole = self.m_flow_borehole
        else:
            self.m_flow_borehole = m_flow_borehole

        # if the mass flow rate has changed, then update it and use new value
        if fluid is None:
            fluid = self.fluid
        else:
            self.fluid = fluid

        self.compute_resistances()

        # Compute resistances that are stored by pygfunction
        # Inner fluid to inner anulus fluid resistance
        R_ff = self.R_f_in + self.R_p_in + self.R_f_a_in
        # Outer annulus fluid to pipe thermal resistance
        R_fp = self.R_p_out + self.R_f_a_out
        self.update_thermal_resistances(R_ff, R_fp)

        # Note: This is the local borehole resistance
        # # Evaluate borehole thermal resistance
        # R_b = 1 / np.trace(1 / self._Rd)

        # Compute and return effective borehole resistance
        R_b_star = self.effective_borehole_thermal_resistance(
            m_flow_borehole, fluid.cp)

        return R_b_star


def compute_Reynolds(m_flow_pipe, r_in, epsilon, fluid):
    # Hydraulic diameter
    D = 2.*r_in
    # Relative roughness
    E = epsilon / D
    # Fluid velocity
    V_flow = m_flow_pipe / fluid.rho
    A_cs = pi * r_in**2
    V = V_flow / A_cs
    # Reynolds number
    Re = fluid.rho * V * D / fluid.mu
    return Re


def compute_Reynolds_concentric(m_flow_pipe, r_a_in, r_a_out, fluid):
    # Hydraulic diameter and radius for concentric tube annulus region
    D_h = 2 * (r_a_out - r_a_in)
    r_h = D_h / 2
    # Cross-sectional area of the annulus region
    A_c = pi * ((r_a_out ** 2) - (r_a_in ** 2))
    # Volume flow rate
    V_dot = m_flow_pipe / fluid.rho
    # Average velocity
    V = V_dot / A_c
    # Reynolds number
    Re = fluid.rho * V * D_h / fluid.mu
    return Re



from . import peak_load_analysis_tool
from . import gfunction
from . import ground_heat_exchangers
from . import coordinates
from . import utilities
from . import domains
from . import search_routines
from . import feature_recognition
from . import media
from . import design
